"""
This module defines the ``happi`` command line utility
"""

import argparse
import fnmatch
import json
import logging
import os
import sys

import coloredlogs
from .utils import is_a_range

import happi

logger = logging.getLogger(__name__)


def get_parser():
    """
    Defines HAPPI shell commands
    """
    # Argument Parser Setup
    parser = argparse.ArgumentParser(description='happi command line tool')

    # Optional args general to all happi operations
    parser.add_argument('--path', type=str,
                        help='path to happi configuration file')
    parser.add_argument('--verbose', '-v', action='store_true',
                        help='Show the degub logging stream')
    parser.add_argument('--version', '-V', action='store_true',
                        help='Current version and location '
                        'of Happi installation.')
    # Subparser to trigger search arguments
    subparsers = parser.add_subparsers(help='Subparsers to search, add, edit',
                                       dest='cmd')
    parser_search = subparsers.add_parser('search', help='Search the happi '
                                          'database')
    parser_search.add_argument('--json', action='store_true',
                               help='Show results as JSON')
    parser_search.add_argument('search_criteria', nargs='+',
                               help='Search criteria: '
                               'field=value. If field= is '
                               'omitted, it will be assumed to be "name". '
                               'You may include as many search criteria as '
                               'you like.')
    parser_add = subparsers.add_parser('add', help='Add new entries')
    parser_add.add_argument('--clone', default='',
                            help='Name of device to use for default parameters'
                            )
    parser_edit = subparsers.add_parser('edit', help='Change existing entry')
    parser_edit.add_argument('name', help='Device to edit')
    parser_edit.add_argument('edits', nargs='+',
                             help='Edits of the form field=value')
    parser_load = subparsers.add_parser('load',
                                        help='Open IPython terminal with '
                                        'device loaded')
    parser_load.add_argument('device_names', nargs='+',
                             help='Devices to load')
    return parser


def happi_cli(args):
    parser = get_parser()
    # print happi usage if no arguments are provided
    if not args:
        parser.print_usage()
        return
    args = parser.parse_args(args)

    # Logging Level handling
    if args.verbose:
        shown_logger = logging.getLogger()
        level = "DEBUG"
    else:
        shown_logger = logging.getLogger('happi')
        level = "INFO"
    coloredlogs.install(level=level, logger=shown_logger,
                        fmt='[%(asctime)s] - %(levelname)s -  %(message)s')
    logger.debug("Set logging level of %r to %r", shown_logger.name, level)

    # Version endpoint
    if args.version:
        print(f'Happi: Version {happi.__version__} from {happi.__file__}')
        return
    logger.debug('Command line arguments: %r' % args)

    client = happi.client.Client.from_config(cfg=args.path)
    logger.debug("Happi client: %r" % client)
    logger.debug('Happi command: %r' % args.cmd)

    if args.cmd == 'search':
        logger.debug("We're in the search block")

        # Get search criteria into dictionary for use by client
        client_args = {}
        range_list = []
        regex_list = []
        is_range = False
        for user_arg in args.search_criteria:
            is_range = False
            if '=' in user_arg:
                criteria, value = user_arg.split('=', 1)
            else:
                criteria = 'name'
                value = user_arg
            if criteria in client_args:
                logger.error(
                    'Received duplicate search criteria %s=%r (was %r)',
                    criteria, value, client_args[criteria]
                )
                return
            if value.replace('.', '').isnumeric():
                logger.debug('Changed %s to float', value)
                value = str(float(value))

            if is_a_range(value):
                start, stop = value.split(',')
                start = float(start)
                stop = float(stop)
                is_range = True
                if start < stop:
                    range_list = client.search_range(criteria, start, stop)
                else:
                    logger.error('Invalid range, make sure start < stop')

            # skip the criteria for range values
            # it won't be a valid criteria for search_regex()
            if is_range:
                pass
            else:
                client_args[criteria] = fnmatch.translate(value)

        regex_list = client.search_regex(**client_args)
        results = regex_list + range_list

        # find the repeated items
        res_size = len(results)
        repeated = []
        for i in range(res_size):
            k = i + 1
            for j in range(k, res_size):
                if results[i] == results[j] and results[i] not in repeated:
                    repeated.append(results[i])

        # we only want to return the ones that have been repeated when
        # they have been matched with both search_regex() & search_range()
        if repeated:
            final_results = repeated
        elif regex_list and not is_range:
            # only matched with search_regex()
            final_results = regex_list
        elif range_list and is_range:
            # only matched with search_range()
            final_results = range_list
        else:
            final_results = []

        if args.json:
            json.dump([dict(res.item) for res in final_results], indent=2,
                      fp=sys.stdout)
        else:
            for res in final_results:
                res.item.show_info()

        if not final_results:
            logger.error('No devices found')
        return final_results
    elif args.cmd == 'add':
        logger.debug('Starting interactive add')
        registry = happi.containers.registry
        if args.clone:
            clone_source = client.find_device(name=args.clone)
            # Must use the same container if cloning
            response = registry.entry_for_class(clone_source.__class__)
        else:
            # Keep Device at registry for backwards compatibility but filter
            # it out of new devices options
            options = os.linesep.join(
                [k for k, _ in registry.items() if k != "Device"]
            )
            logger.info(
                'Please select a container, or press enter for generic '
                'Ophyd Device container: %s%s', os.linesep, options
            )
            response = input()
            if response and response not in registry:
                logger.info('Invalid device container f{response}')
                return
            elif not response:
                response = 'OphydItem'

        container = registry[response]
        kwargs = {}
        for info in container.entry_info:
            valid_value = False
            while not valid_value:
                if args.clone:
                    default = getattr(clone_source, info.key)
                else:
                    default = info.default
                logger.info(f'Enter value for {info.key}, default={default}, '
                            f'enforce={info.enforce}')
                item_value = input()
                if not item_value:
                    if info.optional or args.clone:
                        logger.info(f'Selecting default value {default}')
                        item_value = default
                    else:
                        logger.info('Not an optional field!')
                        continue
                try:
                    info.enforce_value(item_value)
                    valid_value = True
                    kwargs[info.key] = item_value
                except Exception:
                    logger.info(f'Invalid value {item_value}')

        device = client.create_device(container, **kwargs)
        logger.info('Please confirm the following info is correct:')
        device.show_info()
        ok = input('y/N\n')
        if 'y' in ok:
            logger.info('Adding device')
            device.save()
        else:
            logger.info('Aborting')
    elif args.cmd == 'edit':
        logger.debug('Starting edit block')
        device = client.find_device(name=args.name)
        is_invalid_field = False
        for edit in args.edits:
            field, value = edit.split('=', 1)
            try:
                getattr(device, field)
                logger.info('Setting %s.%s = %s', args.name, field, value)
                setattr(device, field, value)
            except Exception as e:
                is_invalid_field = True
                logger.error('Could not edit %s.%s: %s', args.name, field, e)
        if is_invalid_field:
            sys.exit(1)
        device.save()
        device.show_info()
    elif args.cmd == 'load':
        logger.debug('Starting load block')
        logger.info(f'Creating shell with devices {args.device_names}')
        devices = {}
        for name in args.device_names:
            devices[name] = client.load_device(name=name)

        from IPython import start_ipython  # noqa
        start_ipython(argv=['--quick'], user_ns=devices)


def main():
    """Execute the ``happi_cli`` with command line arguments"""
    happi_cli(sys.argv[1:])



import logging
import inspect
import entrypoints

from .item import HappiItem, OphydItem
from .device import Device

logger = logging.getLogger(__name__)

HAPPI_ENTRY_POINT_KEY = "happi.containers"


DEFAULT_REGISTRY = {'Device': Device,
                    'OphydItem': OphydItem,
                    'HappiItem': HappiItem}


class HappiRegistry:
    __instance = None

    def __init__(self):
        if self.__initialized:
            return
        self._registry = {}
        self._reverse_registry = {}
        self.load()
        self.__initialized = True

    def __new__(cls, *args, **kwargs):
        if cls.__instance is None:
            cls.__instance = object.__new__(HappiRegistry)
            cls.__instance.__initialized = False
        return cls.__instance

    def __getitem__(self, item):
        return self._registry.get(item)

    def __contains__(self, item):
        return item in self._registry

    def items(self):
        yield from self._registry.items()

    def entry_for_class(self, klass):
        return self._reverse_registry.get(klass)

    def _safe_add(self, entry_name, klass):
        """
        Add and entry into the registry and raise RuntimeError in case a
        duplicated entry is found.

        Parameters
        ----------
        entry_name : str
            The entrypoint identifier
        klass : class
            The class to add

        Raises
        ------
        RuntimeError
            Raises this exception if the entry is duplicated.
        """
        def make_entry_name():
            """
            Cut out the package name and replace with the entrypoint name so it
            can later be moved if needed.

            Parameters
            ----------
            klass : class

            Returns
            -------
            str
                The formatted name.
            """
            module = ".".join(klass.__module__.split(".")[1:])
            return ".".join([entry_name, module, klass.__name__])

        key = make_entry_name()
        if key in self._registry:
            raise RuntimeError(f"Duplicated entry found for key: {key} "
                               f"and class: {klass}")
        if klass in self._reverse_registry:
            dup_key = self._reverse_registry.get(klass)
            raise RuntimeError(f"Duplicated entry found. Keys: {key} "
                               f"and {dup_key} point to same class: {klass}")

        self._registry[key] = klass
        self._reverse_registry[klass] = key

    def load(self):
        """
        Load entries into the Registry.
        """
        def valid_entry(klass):
            # Avoid happi internal classes due to imports
            return inspect.isclass(klass) and issubclass(klass, HappiItem) \
                   and not klass.__module__.startswith('happi.')

        self._registry = {k: v for k, v in DEFAULT_REGISTRY.items()}
        self._reverse_registry = {v: k for k, v in self._registry.items()}
        _entries = entrypoints.get_group_all(HAPPI_ENTRY_POINT_KEY)

        for entry in _entries:
            try:
                entry_name = entry.name
                obj = entry.load()
            except Exception:
                logger.exception("Failed to load happi.containers entry: %s",
                                 entry)
                continue
            if valid_entry(obj):
                self._safe_add(entry_name, obj)
            elif inspect.ismodule(obj):
                for _, var in inspect.getmembers(obj):
                    if valid_entry(var):
                        self._safe_add(entry_name, var)


registry = HappiRegistry()



"""
Functions to instantiate the Python representations of happi Containers
"""
import asyncio
import importlib
import logging
import sys
import types
from functools import partial
from multiprocessing.pool import ThreadPool

from jinja2 import Environment, meta

from .utils import create_alias

logger = logging.getLogger(__name__)

cache = dict()
main_event_loop = None


def fill_template(template, device, enforce_type=False):
    """
    Fill a Jinja2 template using information from a device

    Parameters
    ----------
    template : str
        Jinja2 template

    device : happi.Device
        Any device container

    enforce_type : bool, optional
        Force the output of the rendered template to match the enforced type of
        the happi information that was used to fill it.
    """
    # Create a template and render our happi information inside it
    env = Environment().from_string(template)
    filled = env.render(**device.post())
    # Find which variable we used in the template, get the type and convert
    # our rendered template to agree with this
    info = meta.find_undeclared_variables(env.environment.parse(template))
    if len(info) == 1 and enforce_type:
        # Get the original attribute back from the device. If this does not
        # exist there is a possibility it is a piece of metadata e.t.c
        try:
            attr_name = info.pop()
            typed_attr = getattr(device, attr_name)
        except AttributeError:
            logger.warning("Can not enforce type to match attribute %s",
                           attr_name)
            return filled
        # If this was a straight substitution with nothing else in the template
        # we can just return the attribute itself thus preserving type
        if str(typed_attr) == filled:
            filled = typed_attr
        # If there is something more complex going on we can attempt to convert
        # it to match the type of the original
        else:
            attr_type = type(typed_attr)
            try:
                filled = attr_type(filled)
            except ValueError:
                logger.exception("Unable to convert %s to %s",
                                 filled, attr_type)
    return filled


def from_container(device, attach_md=True, use_cache=True, threaded=False):
    """
    Load a device from a happi container

    The container is queried for the device_class, args and kwargs. Then if the
    associated package is not already loaded it is imported. The specified
    class is then instantiated with the given args and kwargs provided.

    This function does not attempt to catch exceptions either during module
    imports or device creation. If you would like a series of independent
    devices to be loaded use :func:`.load_devices`.

    By default, the instantiated object has the original container added on as
    ``.md``. This allows applications to utilize additional metadata
    information that may not be included in the basic class constructor. On
    later calls, the container you request is checked against this stored
    metadata. If a discrepancy is found the object is **forced** to reload, not
    retrieved from the cache.

    Parameters
    ----------
    device : happi.Device

    attach_md: bool, optional
        Attach the container to the instantiated object as `md`

    use_cache: bool, optional
        When devices are loaded they are stored in the ``happi.cache``
        dictionary. This means that repeated attempts to load the device will
        return the same object. This prevents unnecessary EPICS connections
        from being initialized in the same process. If a new object is
        needed, set `use_cache` to False and a new object will be created,
        overriding the current cached object. An object with matching name
        and differing metadata will always return a new instantiation of the
        device.

    threaded: bool, optional
        Set this to True when calling inside a thread.

    Returns
    -------
    obj : happi.Device.device_class
    """
    # Return a cached version of the device if present and not forced
    if use_cache and device.name in cache:
        cached_device = cache[device.name]
        # If the metadata has not been modified or we can't review it.
        # Return the cached object
        if hasattr(cached_device, 'md') and cached_device.md == device:
            logger.debug("Loading %s from cache ...", device.name)
            return cached_device
        # Otherwise reload
        else:
            logger.warning("Device %s has already been loaded, but the "
                           "database information has been modified. "
                           "Reloading ...", device.name)

    # Find the class and module of the container.
    if not device.device_class:
        raise ValueError("Device %s does not have an associated Python class",
                         device.name)

    cls = import_class(device.device_class)

    # Create correctly typed arguments from happi information
    def create_arg(arg):
        if not isinstance(arg, str):
            return arg
        return fill_template(arg, device, enforce_type=True)

    # Treat all our args and kwargs as templates
    args = [create_arg(arg) for arg in device.args]
    kwargs = dict((key, create_arg(val))
                  for key, val in device.kwargs.items())
    # Return the instantiated device
    obj = cls(*args, **kwargs)
    # Attach the metadata to the object
    if attach_md:
        try:
            setattr(obj, 'md', device)
        except Exception:
            logger.warning("Unable to attach metadata dictionary to device")

    # Store the device in the cache
    cache[device.name] = obj
    return obj


def import_class(device_class):
    """
    Interpret a device class import string and extract the class object.

    Parameters
    ----------
    device_class : str
        The module path to find the class e.g.
        ``"pcdsdevices.device_types.IPM"``

    Returns
    -------
    cls : type
        The class referred to by the input string.
    """
    mod, cls = device_class.rsplit('.', 1)
    # Import the module if not already present
    # Otherwise use the stashed version in sys.modules
    if mod in sys.modules:
        logger.debug("Using previously imported version of %s", mod)
        mod = sys.modules[mod]
    else:
        logger.debug("Importing %s", mod)
        mod = importlib.import_module(mod)
    # Gather our device class from the given module
    try:
        return getattr(mod, cls)
    except AttributeError as exc:
        raise ImportError("Unable to import %s from %s" %
                          (cls, mod.__name__)) from exc


def load_devices(*devices, pprint=False, namespace=None, use_cache=True,
                 threaded=False, post_load=None, **kwargs):
    """
    Load a series of devices into a namespace

    Parameters
    ----------
    *devices :
        List of happi containers to load

    pprint: bool, optional
        Print results of device loads

    namespace : object, optional
        Namespace to collect loaded devices in. By default this will be a
        ``types.SimpleNamespace``

    use_cache : bool, optional
        If set to ``False``, we'll ignore the cache and always make new
        devices.

    threaded : bool, optional
        Set to True to create each device in a background thread.  Note that
        this assumes that no two devices provided are the same. You are not
        guaranteed to load from the cache correctly if you ask for the same
        device to be loaded twice in the same threaded load.

    post_load : function, optional
        Function of one argument to run on each device after instantiation.
        This is your opportunity to check for good device health during the
        threaded load.

    kwargs:
        Are passed to :func:`.from_container`
    """
    # Create our namespace if we were not given one
    namespace = namespace or types.SimpleNamespace()
    name_list = [container.name for container in devices]
    if threaded:
        # Pre-import because imports in threads have race conditions
        for device in devices:
            try:
                import_class(device.device_class)
            except Exception:
                # Just wait for the normal error handling later
                pass
        global main_event_loop
        if main_event_loop is None:
            main_event_loop = asyncio.get_event_loop()
        pool = ThreadPool(len(devices))
        opt_load = partial(load_device, pprint=pprint, use_cache=use_cache,
                           threaded=True, post_load=post_load, **kwargs)
        loaded_list = pool.map(opt_load, devices)
    else:
        loaded_list = []
        for device in devices:
            loaded = load_device(device, pprint=pprint, use_cache=use_cache,
                                 threaded=False, post_load=post_load, **kwargs)
            loaded_list.append(loaded)
    for dev, name in zip(loaded_list, name_list):
        attr = create_alias(name)
        setattr(namespace, attr, dev)
    return namespace


def load_device(device, pprint=False, threaded=False, post_load=None,
                **kwargs):
    """
    Call :func:`.from_container ` and show success/fail

    Parameters
    ----------
    device : happi.Device

    pprint: bool, optional
        Print results of device loads

    threaded: bool, optional
        Set this to True when calling inside a thread.

    post_load : function, optional
        Function of one argument to run on each device after instantiation.
        This is your opportunity to check for good device health during the
        threaded load.

    kwargs:
        Are passed to :func:`.from_container`
    """
    logger.debug("Loading device %s ...", device.name)

    # We sync with the main thread's loop so that they work as expected later
    if threaded and main_event_loop is not None:
        asyncio.set_event_loop(main_event_loop)

    load_message = "Loading %s [%s] ... "
    success = "\033[32mSUCCESS\033[0m!"
    failed = "\033[31mFAILED\033[0m"
    if pprint:
        device_message = load_message % (device.name, device.device_class)
        if not threaded:
            print(device_message, end='')
    try:
        loaded = from_container(device, **kwargs)
        if post_load is not None:
            post_load(loaded)
        logger.info(load_message + success,
                    device.name, device.device_class)
        if pprint:
            if threaded:
                print(device_message + success)
            else:
                print(success)
    except Exception as exc:
        if pprint:
            if threaded:
                print(device_message + failed)
            else:
                print(failed)
        logger.exception('Error loading %s', device.name)
        loaded = exc
    return loaded



# test_cli.py

import builtins
import logging
import pytest
import happi
import IPython
from happi.cli import happi_cli
from unittest import mock
from happi.errors import SearchError


@pytest.fixture(scope='function')
def happi_cfg(tmp_path, db):
    happi_cfg_path = tmp_path / 'happi.cfg'
    happi_cfg_path.write_text(f"""\
[DEFAULT]'
backend=json
path={db}
""")
    return str(happi_cfg_path.absolute())


@pytest.fixture(scope='function')
def db(tmp_path):
    json_path = tmp_path / 'db.json'
    json_path.write_text("""\
{
    "TST_BASE_PIM": {
        "_id": "TST_BASE_PIM",
        "active": true,
        "args": [
            "{{prefix}}"
        ],
        "beamline": "TST",
        "creation": "Tue Jan 29 09:46:00 2019",
        "device_class": "types.SimpleNamespace",
        "kwargs": {
            "name": "{{name}}"
        },
        "last_edit": "Thu Apr 12 14:40:08 2018",
        "macros": null,
        "name": "tst_base_pim",
        "parent": null,
        "prefix": "TST:BASE:PIM",
        "screen": null,
        "stand": "BAS",
        "system": "diagnostic",
        "type": "OphydItem",
        "z": 3.0
    },
    "TST_BASE_PIM2": {
        "_id": "TST_BASE_PIM2",
        "active": true,
        "args": [
            "{{prefix}}"
        ],
        "beamline": "TST",
        "creation": "Wed Jan 30 09:46:00 2019",
        "device_class": "types.SimpleNamespace",
        "kwargs": {
            "name": "{{name}}"
        },
        "last_edit": "Fri Apr 13 14:40:08 2018",
        "macros": null,
        "name": "tst_base_pim2",
        "parent": null,
        "prefix": "TST:BASE:PIM2",
        "screen": null,
        "stand": "BAS",
        "system": "diagnostic",
        "type": "OphydItem",
        "z": 6.0
    }
}
""")
    return str(json_path.absolute())


def test_cli_version(capsys):
    happi_cli(['--version'])
    readout = capsys.readouterr()
    assert happi.__version__ in readout.out
    assert happi.__file__ in readout.out


def test_cli_no_argument(capsys):
    happi.cli.happi_cli([])
    readout = capsys.readouterr()
    assert 'usage:' in readout.out


def test_search(happi_cfg):
    client = happi.client.Client.from_config(cfg=happi_cfg)
    res = client.search_regex(beamline="TST")
    res_cli = happi.cli.happi_cli(['--verbose', '--path', happi_cfg, 'search',
                                   'beamline=TST'])
    assert [r.device for r in res] == [r.device for r in res_cli]


def test_search_with_name(happi_cfg, caplog):
    client = happi.client.Client.from_config(cfg=happi_cfg)
    res = client.search_regex(name='TST_BASE_PIM2')
    res_cli = happi.cli.happi_cli(['--verbose', '--path', happi_cfg, 'search',
                                   'TST_BASE_PIM2'])
    assert [r.device for r in res] == [r.device for r in res_cli]
    # test duplicate search parameters
    with caplog.at_level(logging.ERROR):
        res_cli = happi.cli.happi_cli(['--verbose', '--path', happi_cfg,
                                       'search', 'TST_BASE_PIM2',
                                       'name=TST_BASE_PIM2'])
        assert "Received duplicate search criteria" in caplog.text
        caplog.clear()


def test_search_z(happi_cfg):
    client = happi.client.Client.from_config(cfg=happi_cfg)
    res = client.search_regex(z="6.0")
    res_cli = happi.cli.happi_cli(['--verbose', '--path', happi_cfg, 'search',
                                   'z=6.0'])
    assert [r.device for r in res] == [r.device for r in res_cli]


def test_search_z_range(happi_cfg, caplog):
    client = happi.client.Client.from_config(cfg=happi_cfg)
    res = client.search_range('z', 3.0, 6.0)
    res_cli = happi.cli.happi_cli(['--verbose', '--path', happi_cfg, 'search',
                                   'z=3.0,6.0'])
    assert [r.device for r in res] == [r.device for r in res_cli]
    # test invalid range
    with caplog.at_level(logging.ERROR):
        res_cli = happi.cli.happi_cli(['--verbose', '--path', happi_cfg,
                                       'search', 'z=6.0,3.0'])
        assert "Invalid range, make sure start < stop" in caplog.text
        caplog.clear()


def test_both_range_and_regex_search(happi_cfg):
    client = happi.client.Client.from_config(cfg=happi_cfg)
    # we're only interested in getting this entry (TST_BASE_PIM2)
    res = client.search_regex(z='6.0')
    # we're going to search for z=3,7 name=TST_BASE_PIM2
    # we should only get in return one entry, even though the z value found 2
    res_cli = happi.cli.happi_cli(['--verbose', '--path', happi_cfg, 'search',
                                   'name=TST_BASE_PIM2', 'z=3.0,7.0'])
    assert [r.device for r in res] == [r.device for r in res_cli]


@pytest.mark.parametrize("from_user, expected_output", [
    # Test add item - succeeding
    pytest.param(['HappiItem', 'happi_name', 'device_class', ['arg1', 'arg2'],
                  {'name': 'my_name'}, True, 'docs', 'y'], [
        "Please select a container, or press enter for generic "
        "Ophyd Device container",
        "Enter value for name, default=None, "
        "enforce=re.compile('[a-z][a-z\\\\_0-9]{2,78}$')",
        "Enter value for device_class, default=None, enforce=<class 'str'>",
        "Enter value for args, default=[], enforce=<class 'list'>",
        "Enter value for kwargs, default={}, enforce=<class 'dict'>",
        "Enter value for active, default=True, enforce=<class 'bool'>",
        "Enter value for documentation, default=None, enforce=<class 'str'>",
        "Please confirm the following info is correct:",
        "Adding device",
        "Storing device HappiItem (name=happi_name) ...",
        "Adding / Modifying information for happi_name ...",
        "HappiItem HappiItem (name=happi_name) has been "
        "succesfully added to the database"],
    ),
    # Test add item - aborting
    pytest.param(['HappiItem', 'happi_name2', 'device_class', ['arg1', 'arg2'],
                  {'name': 'my_name'}, True, 'docs', 'N'], [
        "Please select a container, or press enter for generic "
        "Ophyd Device container",
        "Enter value for name, default=None, "
        "enforce=re.compile('[a-z][a-z\\\\_0-9]{2,78}$')",
        "Enter value for device_class, default=None, enforce=<class 'str'>",
        "Enter value for args, default=[], enforce=<class 'list'>",
        "Enter value for kwargs, default={}, enforce=<class 'dict'>",
        "Enter value for active, default=True, enforce=<class 'bool'>",
        "Enter value for documentation, default=None, enforce=<class 'str'>",
        "Please confirm the following info is correct:",
        "Aborting"],
    ),
    # Test add item - invalid container
    pytest.param(['HappiInvalidItem'], [
        "Please select a container, or press enter for generic "
        "Ophyd Device container",
        "Invalid device container"],
    ),
    # Test add item - no reponse, not an optional field,
    # invalid value, add OphydItem
    pytest.param(['', 7, 'ophyd_name', 'device_class', ['arg1', 'arg2'],
                  {'name': 'my_name'}, True, 'docs', '', 'some_prefix', 'y'], [
        "Please select a container, or press enter for generic "
        "Ophyd Device container",
        "Enter value for name, default=None, "
        "enforce=re.compile('[a-z][a-z\\\\_0-9]{2,78}$')",
        "Invalid value",
        "Enter value for name, default=None, "
        "enforce=re.compile('[a-z][a-z\\\\_0-9]{2,78}$')",
        "Enter value for device_class, default=None, enforce=<class 'str'>",
        "Enter value for args, default=['{{prefix}}'], enforce=<class 'list'>",
        "Enter value for kwargs, default={'name': '{{name}}'}, "
        "enforce=<class 'dict'>",
        "Enter value for active, default=True, enforce=<class 'bool'>",
        "Enter value for documentation, default=None, enforce=<class 'str'>",
        "Enter value for prefix, default=None, enforce=<class 'str'>",
        "Not an optional field!",
        "Enter value for prefix, default=None, enforce=<class 'str'>",
        "Please confirm the following info is correct:",
        "Adding device",
        "Storing device OphydItem (name=ophyd_name) ...",
        "Adding / Modifying information for ophyd_name ...",
        "HappiItem OphydItem (name=ophyd_name) has been "
        "succesfully added to the database"],
    ),
    ], ids=["add_succeeding", "add_aborting",
            "add_invalid_container", "add_not_optional_field"])
def test_add_cli(from_user, expected_output, caplog, happi_cfg):
    with mock.patch.object(builtins, 'input', lambda x=None: from_user.pop(0)):
        happi.cli.happi_cli(['--verbose', '--path', happi_cfg, 'add'])
        caplog.clear()
        for message, expected in zip(caplog.messages, expected_output):
            assert expected in message


@pytest.mark.parametrize("from_user, expected_output", [
    # Test add --clone item - succeeding
    pytest.param(['happi_new_name', 'device_class', ['arg1', 'arg2'],
                  {'name': 'my_name'}, True, '', 'y'], [
        "Enter value for name, default=happi_name, "
        "enforce=re.compile('[a-z][a-z\\\\_0-9]{2,78}$')",
        "Enter value for device_class, default=device_class, "
        "enforce=<class 'str'>",
        "Enter value for args, default=['arg1', 'arg2'], "
        "enforce=<class 'list'>",
        "Enter value for kwargs, default={'name': 'my_name'}, "
        "enforce=<class 'dict'>",
        "Enter value for active, default=True, enforce=<class 'bool'>",
        "Enter value for documentation, default=docs, enforce=<class 'str'>",
        "Selecting default value docs",
        "Please confirm the following info is correct:",
        "Adding device",
        "Storing device HappiItem (name=happi_new_name) ...",
        "Adding / Modifying information for happi_new_name ...",
        "HappiItem HappiItem (name=happi_new_name) has been "
        "succesfully added to the database"], id="clone_succeeding",
    )])
def test_add_clone(from_user, expected_output, caplog, happi_cfg):
    device_info = ['HappiItem', 'happi_name', 'device_class',
                   ['arg1', 'arg2'], {'name': 'my_name'}, True, 'docs', 'y']
    with mock.patch.object(
            builtins, 'input', lambda x=None: device_info.pop(0)):
        # add device first
        happi.cli.happi_cli(['--verbose', '--path', happi_cfg, 'add'])
    with mock.patch.object(builtins, 'input', lambda x=None: from_user.pop(0)):
        # try to clone from previously added device
        happi.cli.happi_cli(
            ['--verbose', '--path', happi_cfg, 'add', '--clone', 'happi_name'])
        caplog.clear()
        for message, expected in zip(caplog.messages, expected_output):
            assert expected in message


def test_add_clone_device_not_fount(happi_cfg):
    with pytest.raises(SearchError):
        happi.cli.happi_cli(
            ['--verbose', '--path', happi_cfg, 'add', '--clone', 'happi_name'])


@pytest.mark.parametrize("from_user, expected_output", [
    # Test edit item name
    pytest.param(['name=new_name'], [
        "Setting happi_name.name = new_name",
        "Saving new entry new_name ...",
        "Removing old entry happi_name ..."], id="edit_name",
    )])
def test_edit(from_user, expected_output, caplog, happi_cfg):
    device_info = ['HappiItem', 'happi_name', 'device_class',
                   ['arg1', 'arg2'], {'name': 'my_name'}, True, 'docs', 'y']
    with mock.patch.object(
            builtins, 'input', lambda x=None: device_info.pop(0)):
        # add device first
        happi.cli.happi_cli(['--verbose', '--path', happi_cfg, 'add'])
    caplog.clear()
    # try edit a previous added device
    happi.cli.happi_cli(
        ['--verbose', '--path', happi_cfg, 'edit',
         'happi_name', from_user.pop(0)])
    for message in caplog.messages:
        for message, expected in zip(caplog.messages, expected_output):
            assert expected in message
    # Test invalid field, note the name is changed to new_name
    with pytest.raises(SystemExit):
        happi.cli.happi_cli(
            ['--verbose', '--path', happi_cfg, 'edit', 'new_name',
             'some_invalid_field=sif'])
    with caplog.at_level(logging.ERROR):
        assert "Could not edit new_name.some_invalid_field: "\
               "'HappiItem' object has no attribute "\
               "'some_invalid_field" in caplog.text


def test_load(caplog, happi_cfg):
    device_info = ['HappiItem', 'happi_name', 'types.SimpleNamespace',
                   [], {'name': 'my_name'}, True, 'docs', 'y']
    with mock.patch.object(
             builtins, 'input', lambda x=None: device_info.pop(0)):
        # add device first
        happi.cli.happi_cli(['--verbose', '--path', happi_cfg, 'add'])
    caplog.clear()
    # try to load the device
    devices = {}
    client = happi.client.Client.from_config(cfg=happi_cfg)
    devices['happi_name'] = client.load_device(name='happi_name')
    with mock.patch.object(IPython, 'start_ipython') as m:
        happi.cli.happi_cli(['--verbose', '--path',
                             happi_cfg, 'load', 'happi_name'])
        m.assert_called_once_with(argv=['--quick'], user_ns=devices)
    with caplog.at_level(logging.INFO):
        assert "Creating shell with devices" in caplog.text



import logging
from unittest.mock import patch

import pytest
import simplejson

from happi import Client, OphydItem
from happi.backends.json_db import JSONBackend

logger = logging.getLogger(__name__)

# Conditional import of pymongo, mongomock
try:
    from mongomock import MongoClient

    from happi.backends.mongo_db import MongoBackend
    supported_backends = ['json', 'mongo']
except ImportError as exc:
    logger.warning('Error importing mongomock : -> %s', exc)
    supported_backends = ['json']

requires_mongo = pytest.mark.skipif('mongo' not in supported_backends,
                                    reason='Missing mongo')

# Conditional import of psdm_qs_cli
try:
    from psdm_qs_cli import QuestionnaireClient

    from happi.backends.qs_db import QSBackend
    has_qs_cli = True
except ImportError as exc:
    logger.warning('Error importing psdm_qs_cli : -> %s', exc)
    has_qs_cli = False


requires_questionnaire = pytest.mark.skipif(not has_qs_cli,
                                            reason='Missing psdm_qs_cli')


try:
    import pcdsdevices  # noqa
    has_pcdsdevices = True
except ImportError as exc:
    logger.warning('error importing pcdsdevices : -> %s', exc)
    has_pcdsdevices = False


requires_pcdsdevices = pytest.mark.skipif(not has_pcdsdevices,
                                          reason='Missing pcdsdevices')


@pytest.fixture(scope='function')
def device_info():
    return {'name': 'alias',
            'z': 400,
            '_id': 'alias',
            'prefix': 'BASE:PV',
            'beamline': 'LCLS',
            'type': 'OphydItem',
            'device_class': 'types.SimpleNamespace',
            'args': list(),
            'kwargs': {'hi': 'oh hello'},
            'location_group': 'LOC',
            'functional_group': 'FUNC',
            }


@pytest.fixture(scope='function')
def device(device_info):
    t = OphydItem(**device_info)
    return t


@pytest.fixture(scope='function')
def valve_info():
    return {'name': 'name',
            'z': 300,
            'prefix': 'BASE:VGC:PV',
            '_id': 'name',
            'beamline': 'LCLS',
            'mps': 'MPS:VGC:PV',
            'location_group': 'LOC',
            'functional_group': 'FUNC',
            }


@pytest.fixture(scope='function')
def valve(valve_info):
    t = OphydItem(**valve_info)
    return t


@pytest.fixture(scope='function')
def mockjsonclient(device_info):
    # Write underlying database
    with open('testing.json', 'w+') as handle:
        simplejson.dump({device_info['name']: device_info},
                        handle)
    # Return handle name
    db = JSONBackend('testing.json')
    return Client(database=db)


@pytest.fixture(scope='function')
@requires_mongo
def mockmongoclient(device_info):
    with patch('happi.backends.mongo_db.MongoClient') as mock_mongo:
        mc = MongoClient()
        mc['test_db'].create_collection('test_collect')
        mock_mongo.return_value = mc
        # Client
        backend = MongoBackend(db='test_db',
                               collection='test_collect')
        client = Client(database=backend)
        # Insert a single device
        client.backend._collection.insert_one(device_info)
        return client


if 'mongo' in supported_backends:
    @pytest.fixture(scope='function', params=supported_backends)
    def happi_client(request, mockmongoclient, mockjsonclient):
        if request.param == 'json':
            return mockjsonclient
        if request.param == 'mongo':
            return mockmongoclient
else:
    @pytest.fixture(scope='function', params=supported_backends)
    def happi_client(request, mockjsonclient):
        return mockjsonclient


@pytest.fixture(scope='module')
def mockqsbackend():
    # Create a very basic mock class
    class MockQuestionnaireClient(QuestionnaireClient):

        def getProposalsListForRun(self, run):
            return {
                'X534': {'Instrument': 'TST', 'proposal_id': 'X534'},
                'LR32': {'Instrument': 'TST', 'proposal_id': 'LR32'},
                'LU34': {'Instrument': 'MFX', 'proposal_id': 'LU34'},
            }

        def getProposalDetailsForRun(self, run_no, proposal):
            return {
                'pcdssetup-motors-1-location': 'Hutch-main experimental',
                'pcdssetup-motors-1-name': 'sam_x',
                'pcdssetup-motors-1-purpose': 'sample x motion',
                'pcdssetup-motors-1-pvbase': 'TST:USR:MMS:01',
                'pcdssetup-motors-1-stageidentity': 'IMS MD23',
                'pcdssetup-motors-2-location': 'Hutch-main experimental',
                'pcdssetup-motors-2-name': 'sam_z',
                'pcdssetup-motors-2-purpose': 'sample z motion',
                'pcdssetup-motors-2-pvbase': 'TST:USR:MMS:02',
                'pcdssetup-motors-2-stageidentity': 'IMS MD23',
                'pcdssetup-motors-3-location': 'Hutch-main experimental',
                'pcdssetup-motors-3-name': 'sam_y',
                'pcdssetup-motors-3-purpose': 'sample y motion',
                'pcdssetup-motors-3-pvbase': 'TST:USR:MMS:03',
                'pcdssetup-motors-3-stageidentity': 'IMS MD32',
                'pcdssetup-motors-4-location': 'Hutch-main experimental',
                'pcdssetup-motors-4-name': 'sam_r',
                'pcdssetup-motors-4-purpose': 'sample rotation',
                'pcdssetup-motors-4-pvbase': 'TST:USR:MMS:04',
                'pcdssetup-motors-4-stageidentity': 'IMS MD23',
                'pcdssetup-motors-5-location': 'Hutch-main experimental',
                'pcdssetup-motors-5-name': 'sam_az',
                'pcdssetup-motors-5-purpose': 'sample azimuth',
                'pcdssetup-motors-5-pvbase': 'TST:USR:MMS:05',
                'pcdssetup-motors-5-stageidentity': 'IMS MD23',
                'pcdssetup-motors-6-location': 'Hutch-main experimental',
                'pcdssetup-motors-6-name': 'sam_flip',
                'pcdssetup-motors-6-purpose': 'sample flip',
                'pcdssetup-motors-6-pvbase': 'TST:USR:MMS:06',
                'pcdssetup-motors-6-stageidentity': 'IMS MD23',
                'pcdssetup-trig-1-delay': '0.00089',
                'pcdssetup-trig-1-eventcode': '198',
                'pcdssetup-trig-1-name': 'Overview_trig',
                'pcdssetup-trig-1-polarity': 'positive',
                'pcdssetup-trig-1-purpose': 'Overview',
                'pcdssetup-trig-1-pvbase': 'MFX:REC:EVR:02:TRIG1',
                'pcdssetup-trig-1-width': '0.00075',
                'pcdssetup-trig-2-delay': '0.000894348',
                'pcdssetup-trig-2-eventcode': '198',
                'pcdssetup-trig-2-name': 'Meniscus_trig',
                'pcdssetup-trig-2-polarity': 'positive',
                'pcdssetup-trig-2-purpose': 'Meniscus',
                'pcdssetup-trig-2-pvbase': 'MFX:REC:EVR:02:TRIG3',
                'pcdssetup-trig-2-width': '0.0005',
                'pcdssetup-ao-1-device': 'Acromag IP231 16-bit',
                'pcdssetup-ao-1-name': 'irLed',
                'pcdssetup-ao-1-purpose': 'IR LED',
                'pcdssetup-ao-2-channel': '6',
                'pcdssetup-ao-2-device': 'Acromag IP231 16-bit',
                'pcdssetup-ao-2-name': 'laser_shutter_opo',
                'pcdssetup-ao-2-purpose': 'OPO Shutter',
                'pcdssetup-ao-3-channel': '7',
                'pcdssetup-ao-3-device': 'Acromag IP231 16-bit',
                'pcdssetup-ao-3-name': 'laser_shutter_evo1',
                'pcdssetup-ao-3-purpose': 'EVO Shutter1',
                'pcdssetup-ao-4-channel': '2',
                'pcdssetup-ao-4-device': 'Acromag IP231 16-bit',
                'pcdssetup-ao-4-name': 'laser_shutter_evo2',
                'pcdssetup-ao-4-purpose': 'EVO Shutter2',
                'pcdssetup-ao-5-channel': '3',
                'pcdssetup-ao-5-device': 'Acromag IP231 16-bit',
                'pcdssetup-ao-5-name': 'laser_shutter_evo3',
                'pcdssetup-ao-5-purpose': 'EVO Shutter3',
                'pcdssetup-ao-1-pvbase': 'MFX:USR:ao1:0',
                'pcdssetup-ao-2-pvbase': 'MFX:USR:ao1:6',
                'pcdssetup-ao-3-pvbase': 'MFX:USR:ao1:7',
                'pcdssetup-ao-4-pvbase': 'MFX:USR:ao1:2',
                'pcdssetup-ao-5-pvbase': 'MFX:USR:ao1:3',
                'pcdssetup-ai-1-device': 'Acromag IP231 16-bit',
                'pcdssetup-ai-1-name': 'irLed',
                'pcdssetup-ai-1-purpose': 'IR LED',
                'pcdssetup-ai-1-pvbase': 'MFX:USR:ai1:0',
            }

        def getExpName2URAWIProposalIDs(self):
            return {
                'tstx53416': 'X534',
                'tstlr3216': 'LR32',
                'mfxlu3417': 'LU34',
            }

    with patch('happi.backends.qs_db.QuestionnaireClient') as qs_cli:
        # Replace QuestionnaireClient with our test version
        mock_qs = MockQuestionnaireClient(use_kerberos=False,
                                          user='user', pw='pw')
        qs_cli.return_value = mock_qs
        # Instantiate a fake device
        backend = QSBackend('tstlr3216')
        return backend


@pytest.fixture(scope='function')
def three_valves(happi_client):
    valve1 = {'name': 'valve1',
              'z': 300,
              'prefix': 'BASE:VGC1:PV',
              '_id': 'VALVE1',
              'beamline': 'LCLS',
              'mps': 'MPS:VGC:PV',
              'type': 'OphydItem',
              'location_group': 'LOC',
              'functional_group': 'FUNC',
              'device_class': 'types.SimpleNamespace',
              'args': list(),
              'kwargs': {'hi': 'oh hello'},
              }

    valve2 = {'name': 'valve2',
              'z': 301,
              'prefix': 'BASE:VGC2:PV',
              '_id': 'VALVE2',
              'beamline': 'LCLS',
              'mps': 'MPS:VGC:PV',
              'type': 'OphydItem',
              'location_group': 'LOC',
              'functional_group': 'FUNC',
              'device_class': 'types.SimpleNamespace',
              'args': list(),
              'kwargs': {'hi': 'oh hello'},
              }

    valve3 = {'name': 'valve3',
              'z': 301,
              'prefix': 'BASE:VGC3:PV',
              '_id': 'VALVE3',
              'beamline': 'LCLS',
              'mps': 'MPS:VGC:PV',
              'location_group': 'LOC',
              'functional_group': 'FUNC',
              'type': 'OphydItem',
              'device_class': 'types.SimpleNamespace',
              'args': list(),
              'kwargs': {'hi': 'oh hello'},
              }

    for dev in happi_client.all_items:
        happi_client.backend.delete(dev['_id'])

    valves = dict(
        VALVE1=valve1,
        VALVE2=valve2,
        VALVE3=valve3,
    )

    for name, valve in valves.items():
        happi_client.backend.save(name, valve, insert=True)
    return valves



import copy
import io
import re

import pytest

from happi import Device
from happi.errors import ContainerError
from happi.device import EntryInfo


def test_get(device, device_info):
    assert device.name == device_info['name']


def test_init(device, device_info):
    assert device.prefix == device_info['prefix']
    assert device.name == device_info['name']


def test_list_enforce():
    # Generic device with a list enforce
    class MyDevice(Device):
        list_attr = EntryInfo(enforce=['a', 'b', 'c'])

    # Make sure we can set without error
    d = MyDevice()
    d.list_attr = 'b'
    assert d.list_attr == 'b'
    # Mase sure we can not set outside the list
    d = MyDevice()
    with pytest.raises(ValueError):
        d.list_attr = 'd'


def test_regex_enforce():
    class MyDevice(Device):
        re_attr = EntryInfo(enforce=re.compile(r'[A-Z]{2}$'))

    d = MyDevice()
    d.re_attr = 'AB'

    d = MyDevice()
    with pytest.raises(ValueError):
        d.re_attr = 'ABC'


def test_set(device):
    device.name = 'new_name'
    assert device.name == 'new_name'


def test_optional(device):
    assert device.documentation is None


def test_enforce(device):
    with pytest.raises(ValueError):
        device.name = 'Invalid!Name'


def test_container_error():
    with pytest.raises(ContainerError):
        class MyDevice(Device):
            fault = EntryInfo(enforce=int,  default='not-int')


def test_mandatory_info(device):
    for info in ('prefix', 'name'):
        assert info in device.mandatory_info


def test_restricted_attr():
    with pytest.raises(TypeError):
        class MyDevice(Device):
            info_names = EntryInfo()


def test_post(device, device_info):
    post = device.post()
    assert post['prefix'] == device_info['prefix']
    assert post['name'] == device_info['name']


def test_show_info(device, device_info):
    f = io.StringIO()
    device.show_info(handle=f)
    f.seek(0)
    out = f.read()
    device_info.pop('_id')
    assert '_id' not in out
    assert all([info in out for info in device_info.keys()])


def test_device_equivalance():
    a = Device(name='abcd', prefix='b')
    b = Device(name='abcd', prefix='b')
    c = Device(name='cbcd', prefix='b')
    assert a == b
    assert not c == a


def test_dictify():
    a = Device(name='abcd', prefix='b')
    assert dict(a) == a.post()


def test_device_copy():
    a = Device(name='abcd', prefix='b')
    b = copy.copy(a)
    assert dict(a) == dict(b)


def test_device_deepcopy():
    a = Device(name='abcd', prefix='abc', kwargs={'abc': 'def'})

    c = copy.deepcopy(a)
    assert a.kwargs == c.kwargs
    assert id(a.kwargs) != id(c.kwargs)


def test_add_and_save(three_valves, device, happi_client):
    device.active = True
    happi_client.add_device(device)
    device.active = False
    device.save()

    assert not happi_client[device.name].device.active



"""
Backend implemenation using simplejson
"""
import contextlib
import os
import os.path
import logging
import math
import re

import simplejson as json

from .core import _Backend
from ..errors import SearchError, DuplicateError

logger = logging.getLogger(__name__)


try:
    import fcntl
except ImportError:
    logger.warning("Unable to import 'fcntl'. Will be unable to lock files")
    fcntl = None


@contextlib.contextmanager
def _load_and_store_context(backend):
    '''
    A context manager to load, and optionally store the JSON database at the
    end
    '''
    db = backend._load_or_initialize()
    yield db
    backend.store(db)


class JSONBackend(_Backend):
    """
    JSON database

    The happi information is kept in a single dictionary large dictionary that
    is stored using `simplejson`

    Parameters
    ----------
    path : str
        Path to JSON file

    initialze : bool, optional
        Initialize a new empty JSON file to begin filling
    """
    def __init__(self, path, initialize=False):
        self.path = path
        # Create a new JSON file if requested
        if initialize:
            self.initialize()

    def _load_or_initialize(self):
        '''
        Load an existing database or initialize a new one.
        '''
        try:
            return self.load()
        except FileNotFoundError:
            logger.debug('Initializing new database')

        self.initialize()
        return self.load()

    @property
    def all_devices(self):
        """
        All of the devices in the database
        """
        json = self._load_or_initialize()
        return list(json.values())

    def initialize(self):
        """
        Initialize a new JSON file database

        Raises
        ------
        PermissionError:
            If the JSON file specified by `path` already exists

        Notes
        -----
        This is exists because the `store` and `load` methods assume that the
        given path already points to a readable JSON file. In order to begin
        filling a new database, an empty but valid JSON file is created
        """
        # Do not overwrite existing databases
        if os.path.exists(self.path) and os.stat(self.path).st_size > 0:
            raise PermissionError("File {} already exists. Can not initialize "
                                  "a new database.".format(self.path))
        # Dump an empty dictionary
        self.store({})

    def load(self):
        """
        Load the JSON database
        """
        with open(self.path, 'r') as f:
            raw_json = f.read()

        # Allow for empty files to be considered valid databases:
        return json.loads(raw_json) if raw_json else {}

    def store(self, db):
        """
        Stache the database back into JSON

        Parameters
        ----------
        db : dict
            Dictionary to store in JSON

        Raises
        ------
        BlockingIOError:
            If the file is already being used by another happi operation
        """
        with open(self.path, 'w+') as f:
            # Create lock in filesystem
            if fcntl is not None:
                fcntl.flock(f, fcntl.LOCK_EX | fcntl.LOCK_NB)
            # Dump to file
            try:
                json.dump(db, f, sort_keys=True, indent=4)

            finally:
                if fcntl is not None:
                    # Release lock in filesystem
                    fcntl.flock(f, fcntl.LOCK_UN)

    def _iterative_compare(self, comparison):
        """
        Yields documents in which ``comparison(name, doc)`` returns True.

        Parameters
        ----------
        comparison : callable
            A comparison function with a signature of (device_id, doc)
        """
        db = self._load_or_initialize()
        if not db:
            return

        for name, doc in db.items():
            try:
                if comparison(name, doc):
                    yield doc
            except Exception as ex:
                logger.debug('Comparison method failed: %s', ex, exc_info=ex)

    def get_by_id(self, id_):
        '''
        Get a device by ID if it exists, or None

        Parameters
        ----------
        id_
        '''
        db = self._load_or_initialize()
        return db.get(id_)

    def find(self, to_match):
        """
        Find an instance or instances that matches the search criteria

        Parameters
        ----------
        to_match : dict
            Requested information, all of which must match
        """
        def comparison(name, doc):
            return all(value == doc[key]
                       for key, value in to_match.items())

        yield from self._iterative_compare(comparison)

    def find_range(self, key, *, start, stop=None, to_match):
        """
        Find an instance or instances that matches the search criteria, such
        that ``start <= entry[key] < stop``.

        Parameters
        ----------
        key : str
            The database key to search

        start : int or float
            Inclusive minimum value to filter ``key`` on

        end : float, optional
            Exclusive maximum value to filter ``key`` on

        to_match : dict
            Requested information, where the values must match exactly
        """
        def comparison(name, doc):
            if all(value == doc[k] for k, value in to_match.items()):
                value = doc.get(key)
                try:
                    return start <= value < stop
                except Exception:
                    ...

        if key in to_match:
            raise ValueError('Cannot specify the same key in `to_match` as '
                             'the key for the range.')
        if stop is None:
            stop = math.inf
        if start >= stop:
            raise ValueError(f"Invalid range: {start} >= {stop}")

        yield from self._iterative_compare(comparison)

    def find_regex(self, to_match, *, flags=re.IGNORECASE):
        """
        Find an instance or instances that matches the search criteria,
        using regular expressions.

        Parameters
        ----------
        to_match : dict
            Requested information, where the values are regular expressions.
        """
        def comparison(name, doc):
            return regexes and all(key in doc and regex.match(str(doc[key]))
                                   for key, regex in regexes.items())

        regexes = {
            key: re.compile(value, flags=flags)
            for key, value in to_match.items()
        }

        yield from self._iterative_compare(comparison)

    def save(self, _id, post, insert=True):
        """
        Save information to the database

        Parameters
        ----------
        _id : str
            ID of device

        post : dict
            Information to place in database

        insert : bool, optional
            Whether or not this a new device to the database

        Raises
        ------
        DuplicateError:
            If insert is True, but there is already a device with the provided
            _id

        SearchError:
            If insert is False, but there is no device with the provided _id

        PermissionError:
            If the write operation fails due to issues with permissions
        """
        with _load_and_store_context(self) as db:
            # New device
            if insert:
                if _id in db.keys():
                    raise DuplicateError(f"Device {_id} already exists")
                # Add _id keyword
                post.update({'_id': _id})
                # Add to database
                db[_id] = post
            # Updating device
            else:
                # Edit information
                try:
                    db[_id].update(post)
                except KeyError:
                    raise SearchError("No device found {}".format(_id))

    def delete(self, _id):
        """
        Delete a device instance from the database

        Parameters
        ----------
        _id : str
            ID of device

        Raises
        ------
        PermissionError:
            If the write operation fails due to issues with permissions
        """
        with _load_and_store_context(self) as db:
            try:
                db.pop(_id)
            except KeyError:
                raise SearchError(
                    f'ID not found in database: {_id!r}'
                ) from None



"""
Basic module utilities
"""
import logging

logger = logging.getLogger(__name__)


def create_alias(name):
    """
    Clean an alias to be an acceptable Python variable
    """
    return name.replace(' ', '_').replace('.', '_').lower()


def get_happi_entry_value(entry, key, search_extraneous=True):
    extraneous = entry.extraneous
    value = getattr(entry, key, None)
    if value is None and search_extraneous:
        # Try to look at extraneous
        value = extraneous.get(key, None)

    if not value:
        raise ValueError('Invalid Key for Device.')
    return value


def is_number(str_value):
    """
    Checks if it is a valid float number
    """
    try:
        float(str_value)
        return True
    except ValueError:
        return False


def is_a_range(str_value):
    """
    Checks to see if it is a range.
    It needs to have two valid values separated by a comma.
    Ex:
        1,100
        -2,8
        100,2
    """
    if ',' in str_value:
        start, stop = str_value.split(',')
        if is_number(start) and is_number(stop):
            return True
        else:
            logger.error("Possibly provided invalid numbers for a range")
            return False
    else:
        return False



import os
import logging
import re
import tempfile
import types

import pytest

from happi import Client, OphydItem
from happi.backends.json_db import JSONBackend
from happi.errors import SearchError, DuplicateError, EntryError

logger = logging.getLogger(__name__)


@pytest.fixture(scope='function')
def xdg_config_home(tmp_path):
    config_home = tmp_path / 'xdg_config_home'
    config_home.mkdir()
    return config_home


@pytest.fixture(scope='function')
def happi_cfg(tmp_path, xdg_config_home):
    # Store current happi config
    xdg_cfg = os.environ.get("XDG_CONFIG_HOME", '')
    happi_cfg = os.environ.get("HAPPI_CFG", '')

    # Setup environment variables
    os.environ['XDG_CONFIG_HOME'] = str(xdg_config_home)
    os.environ['HAPPI_CFG'] = ''

    # Write file
    happi_cfg = xdg_config_home / 'happi.cfg'
    happi_cfg.write_text("""\
[DEFAULT]
backend=json
path=db.json
""")
    yield str(happi_cfg)
    # Restore environment variables
    os.environ["HAPPI_CFG"] = str(happi_cfg)
    os.environ["XDG_CONFIG_HOME"] = xdg_cfg


def test_find_document(happi_client, device_info):
    doc = happi_client.find_document(**device_info)
    assert doc.pop('prefix') == device_info['prefix']
    assert doc.pop('name') == device_info['name']
    # Remove id and check
    prefix = device_info.pop('prefix')
    doc = happi_client.find_document(**device_info)

    assert doc.pop('prefix') == prefix
    assert doc.pop('name') == device_info['name']
    # Not available
    with pytest.raises(SearchError):
        doc = happi_client.find_document(prefix='Does not Exist')


def test_create_device(happi_client, device_info):
    device = happi_client.create_device(OphydItem, **device_info)
    assert device.prefix == device_info['prefix']
    assert device.name == device_info['name']
    # Invalid Entry
    with pytest.raises(TypeError):
        happi_client.create_device(int)


def test_all_devices(happi_client, device):
    assert happi_client.all_devices == [device]


def test_add_device(happi_client, valve):
    happi_client.add_device(valve)
    # No duplicates
    with pytest.raises(DuplicateError):
        happi_client.add_device(valve)
    # No incompletes
    d = OphydItem()
    with pytest.raises(EntryError):
        happi_client.add_device(d)


def test_add_and_find_device(happi_client, valve, valve_info):
    happi_client.add_device(valve)
    loaded_device = happi_client.find_device(**valve_info)
    assert loaded_device.prefix == valve.prefix
    assert loaded_device.name == valve.name


def test_find_device(happi_client, device_info):
    device = happi_client.find_device(**device_info)
    assert isinstance(device, OphydItem)
    assert device.prefix == device_info['prefix']
    assert device.name == device_info['name']
    # Test edit and save
    device.stand = 'DG3'
    device.save()
    loaded_device = happi_client.find_device(**device_info)
    assert loaded_device.prefix == device_info['prefix']
    assert loaded_device.name == device_info['name']
    # Bad load
    bad = {'a': 'b'}
    happi_client.backend.save('a', bad, insert=True)
    with pytest.raises(EntryError):
        happi_client.find_device(**bad)


def test_change_item_name(happi_client, device_info):
    device = happi_client.find_device(**device_info)
    assert device.name != 'new_name'
    device.name = 'new_name'
    device.save()
    # old entry should not exist anymore
    with pytest.raises(SearchError):
        happi_client.find_device(**device_info)
    # new entry with new name should be there
    new_device = happi_client.find_device(name=device.name)
    # prefix or other attributes should be the same
    assert new_device.prefix == device.prefix
    # we should only have one deivce now which is the new one
    assert happi_client.all_items == [new_device]


def test_validate(happi_client):
    # No bad devices
    assert happi_client.validate() == list()
    # A single bad device
    happi_client.backend.save('_id', {happi_client._id_key: 'bad'},
                              insert=True)
    assert happi_client.validate() == ['bad']


def test_search(happi_client, device, valve, device_info, valve_info):
    happi_client.add_device(valve)
    res = happi_client.search(name=device_info['name'])
    # Single search return
    assert len(res) == 1
    loaded_device = res[0].device
    assert loaded_device.prefix == device_info['prefix']
    assert loaded_device.name == device_info['name']
    # No results
    assert not happi_client.search(name='not')
    # Returned as dict
    res = happi_client.search(**device_info)
    loaded_device = res[0].device
    assert loaded_device['prefix'] == device_info['prefix']
    assert loaded_device['name'] == device_info['name']
    # Search off keyword
    res = happi_client.search(beamline='LCLS')
    assert len(res) == 2


def test_search_range(happi_client, device, valve, device_info, valve_info):
    happi_client.add_device(valve)
    # Search between two points
    res = happi_client.search_range('z', start=0, end=500)
    assert len(res) == 2
    # Search between two points, nothing found
    res = happi_client.search_range('z', start=10000, end=500000)
    assert not res
    # Search without an endpoint
    res = happi_client.search_range('z', start=0)
    assert len(res) == 2
    # Search invalid range
    with pytest.raises(ValueError):
        happi_client.search_range('z', start=1000, end=5)


def test_search_regex(happi_client, three_valves):
    def find(**kwargs):
        return [
            dict(item) for item in
            happi_client.search_regex(**kwargs, flags=re.IGNORECASE)
        ]

    valve1 = dict(happi_client['VALVE1'])
    valve2 = dict(happi_client['VALVE2'])
    valve3 = dict(happi_client['VALVE3'])

    assert find(beamline='LCLS') == [valve1, valve2, valve3]
    assert find(beamline='lcls') == [valve1, valve2, valve3]
    assert find(beamline='nomatch') == []
    assert find(_id=r'VALVE\d') == [valve1, valve2, valve3]
    assert find(_id='VALVE[13]') == [valve1, valve3]
    assert find(prefix='BASE:VGC[23]:PV') == [valve2, valve3]


def test_get_by_id(happi_client, device, valve, device_info, valve_info):
    happi_client.add_device(valve)
    name = valve_info['name']
    for k, v in valve_info.items():
        assert happi_client[name][k] == valve_info[k]


def test_remove_device(happi_client, device, valve, device_info):
    happi_client.remove_device(device)
    assert list(happi_client.backend.find(device_info)) == []
    # Invalid Device
    with pytest.raises(ValueError):
        happi_client.remove_device(5)
    # Valve not in dictionary
    with pytest.raises(SearchError):
        happi_client.remove_device(valve)


def test_export(happi_client, valve):
    # Setup client with both devices
    happi_client.add_device(valve)
    fd, temp_path = tempfile.mkstemp()
    happi_client.export(open(temp_path, 'w+'),
                        sep=',',
                        attrs=['name', 'prefix'])
    exp = open(temp_path, 'r').read()
    assert "alias,BASE:PV" in exp
    assert "name,BASE:VGC:PV" in exp
    # Cleanup
    os.remove(temp_path)
    os.close(fd)


def test_load_device(happi_client, device):
    device = happi_client.load_device(name=device.name)
    assert isinstance(device, types.SimpleNamespace)
    assert device.hi == 'oh hello'


def test_find_cfg(happi_cfg):
    # Use our config directory
    assert happi_cfg == Client.find_config()
    # Set the path explicitly using HAPPI env variable
    os.environ['HAPPI_CFG'] = happi_cfg
    assert happi_cfg == Client.find_config()


def test_from_cfg(happi_cfg):
    client = Client.from_config()
    assert isinstance(client.backend, JSONBackend)
    assert client.backend.path == 'db.json'


def test_choices_for_field(happi_client):
    name_choices = happi_client.choices_for_field('name')
    assert name_choices == {'alias'}
    prefix_choices = happi_client.choices_for_field('prefix')
    assert prefix_choices == {'BASE:PV'}
    with pytest.raises(SearchError):
        happi_client.choices_for_field('not_a_field')


def test_searchresults(happi_client, three_valves):
    valve1 = happi_client['VALVE1']
    print(repr(valve1))
    print(dict(valve1))
    assert valve1.metadata == valve1
    assert isinstance(valve1.get(), types.SimpleNamespace)


def test_client_mapping(happi_client, three_valves):
    assert len(happi_client) == 3
    assert list(dict(happi_client)) == ['VALVE1', 'VALVE2', 'VALVE3']
    for name in happi_client:
        assert happi_client[name]['name']



"""
PyMongo Backend Implementation
"""
import logging
import re

import bson.regex

from pymongo import MongoClient
from pymongo.errors import ServerSelectionTimeoutError, OperationFailure

from .core import _Backend
from ..errors import DatabaseError, SearchError, DuplicateError

logger = logging.getLogger(__name__)


class MongoBackend(_Backend):
    """
    Abstraction for MongoDB backend

    Parameters
    ----------
    host : str, optional
        Hostname for MongoDB

    user : str, optional
        Username for MongoDB instance

    pw :str, optional
        Password for given username

    host : str, optional
        Host of the MongoDB instance

    db : str, optional
        Database name within the MongoDB instance

    timeout : float, optional
        Time to wait for connection attempt
    """
    _timeout = 5
    _conn_str = 'mongodb://{user}:{pw}@{host}/{db}'  # String for login

    def __init__(self, host=None, user=None,
                 pw=None, db=None, collection=None,
                 timeout=None):
        # Default timeout
        timeout = timeout or self._timeout
        # Format connection string
        conn_str = self._conn_str.format(user=user, pw=pw,
                                         host=host, db=db)
        logging.debug('Attempting connection using %s ', conn_str)
        self._client = MongoClient(conn_str, serverSelectionTimeoutMS=timeout)
        self._db = self._client[db]
        # Load collection
        try:
            if collection not in self._db.list_collection_names():
                raise DatabaseError('Unable to locate collection {} '
                                    'in database'.format(collection))
            self._collection = self._db[collection]
        # Unable to view collection names
        except OperationFailure as e:
            raise PermissionError(e)
        # Unable to connect to MongoDB instance
        except ServerSelectionTimeoutError:
            raise DatabaseError('Unable to connect to MongoDB instance, check '
                                'that the server is running on the host and '
                                'port specified at startup')

    @property
    def all_devices(self):
        """
        List of all device sub-dictionaries
        """
        return self._collection.find()

    def find(self, to_match):
        """
        Yield all instances that match the given search criteria

        Parameters
        ----------
        to_match : dict
            Requested information
        """
        yield from self._collection.find(to_match)

    def find_range(self, key, *, start, stop=None, to_match):
        """
        Find an instance or instances that matches the search criteria, such
        that ``start <= entry[key] < stop``.

        Parameters
        ----------
        key : str
            The database key to search

        start : int or float
            Inclusive minimum value to filter ``key`` on

        end : float, optional
            Exclusive maximum value to filter ``key`` on

        to_match : dict
            Requested information, where the values must match exactly
        """
        if key in to_match:
            raise ValueError('Cannot specify the same key in `to_match` as '
                             'the key for the range.')

        match = {key: {'$gte': start}}
        if stop is not None:
            if start >= stop:
                raise ValueError(f"Invalid range: {start} >= {stop}")
            match[key]['$lt'] = stop

        match.update(**to_match)
        yield from self._collection.find(match)

    def get_by_id(self, _id):
        """
        Get a device by ID if it exists, or None

        Parameters
        ----------
        _id : str
            The device ID
        """
        for item in self._collection.find({'_id': _id}):
            return item

    def find_regex(self, to_match, *, flags=re.IGNORECASE):
        """
        Yield all instances that match the given search criteria

        Parameters
        ----------
        to_match : dict
            Requested information, with each value being a regular expression
        """
        regexes = {}
        for key, value in to_match.items():
            try:
                reg = re.compile(value, flags=flags)
                regexes[key] = bson.regex.Regex.from_native(reg)
            except Exception as ex:
                raise ValueError(f'Failed to create regular expression from '
                                 f'{key}={value!r}: {ex}') from ex

        yield from self._collection.find(regexes)

    def save(self, _id, post, insert=True):
        """
        Save information to the database

        Parameters
        ----------
        _id : str
            ID of device

        post : dict
            Information to place in database

        insert : bool, optional
            Whether or not this a new device to the database

        Raises
        ------
        DuplicateError:
            If insert is True, but there is already a device with the provided
            _id

        SearchError:
            If insert is False, but there is no device with the provided _id

        PermissionError:
            If the write operation fails due to issues with permissions
        """
        try:
            # Add to database
            result = self._collection.update_one({'_id': _id},
                                                 {'$set': post},
                                                 upsert=insert)
        except OperationFailure:
            raise PermissionError("Unauthorized command, make sure you are "
                                  "using a user with write permissions")

        if insert and not result.upserted_id:
            raise DuplicateError('Device with id {} has already been entered '
                                 'into the database, use load_device and '
                                 'save if you wish to make changes to the '
                                 'device'.format(_id))

        if not insert and result.matched_count == 0:
            raise SearchError('No device found with id {} please, if this is '
                              'a new device, try add_device. If not, make '
                              'sure that the device information being sent is '
                              'correct'.format(_id))

    def delete(self, _id):
        """
        Delete a device instance from the database

        Parameters
        ----------
        _id : str
            ID of device
        """
        res = self._collection.delete_one({'_id': _id})
        if res.deleted_count < 1:
            raise SearchError(f'ID not found in database: {_id!r}')



import fcntl
import os
import os.path

import pytest
import simplejson

from happi import Client
from happi.backends.json_db import JSONBackend
from happi.errors import DuplicateError, SearchError

from .conftest import (requires_mongo, requires_pcdsdevices,
                       requires_questionnaire)


@pytest.fixture(scope='function')
def mockmongo(mockmongoclient):
    return mockmongoclient.backend


@pytest.fixture(scope='function')
def mockjson(device_info, valve_info):
    # Write underlying database
    with open('testing.json', 'w+') as handle:
        simplejson.dump({device_info['_id']: device_info},
                        handle)
    # Return handle name
    yield JSONBackend('testing.json')

    # Delete file
    os.remove('testing.json')


@requires_mongo
def test_mongo_find(valve_info, device_info, mockmongo):
    mm = mockmongo
    mm._collection.insert_one(valve_info)

    def find(**kwargs):
        return list(mm.find(kwargs))

    assert find(beamline='BLERG') == []
    # Single device by id
    assert [device_info] == find(_id=device_info['_id'])
    # Single device by kwarg
    assert [valve_info] == find(prefix=valve_info['prefix'])
    # No multiple devices expected
    assert find(beamline='BLERG') == []
    # Multiple devices by id
    assert [device_info] == find(_id=device_info['_id'])
    # Multiple devices by kwarg
    assert [device_info] == find(prefix=device_info['prefix'])
    # Multiple devices expected
    assert all(info in find(beamline='LCLS')
               for info in (device_info, valve_info))


@requires_mongo
def test_mongo_save(mockmongo, device_info, valve_info):
    # Duplicate device
    with pytest.raises(DuplicateError):
        mockmongo.save(device_info[Client._id_key], device_info, insert=True)

    # Device not found
    with pytest.raises(SearchError):
        mockmongo.save(valve_info[Client._id_key], valve_info, insert=False)

    # Add to database
    mockmongo.save(valve_info[Client._id_key], valve_info, insert=True)
    assert mockmongo._collection.find_one(valve_info) == valve_info


@requires_mongo
def test_mongo_delete(mockmongo, device_info):
    mockmongo.delete(device_info[Client._id_key])
    assert mockmongo._collection.find_one(device_info) is None


def test_json_find(valve_info, device_info, mockjson):
    mm = mockjson
    # Write underlying database
    with open(mm.path, 'w+') as handle:
        simplejson.dump({valve_info['_id']: valve_info,
                         device_info['_id']: device_info},
                        handle)

    def find(**kwargs):
        return list(mm.find(kwargs))

    # No single device expected
    assert find(beamline='BLERG') == []
    # Single device by id
    assert [device_info] == find(_id=device_info['_id'])
    # Single device by kwarg
    assert [valve_info] == find(prefix=valve_info['prefix'])
    # No multiple devices expected
    assert find(beamline='BLERG') == []
    # Multiple devices by id
    assert [device_info] == find(_id=device_info['_id'])
    # Multiple devices by kwarg
    assert [device_info] == find(prefix=device_info['prefix'])
    # Multiple devices expected
    assert all(info in find(beamline='LCLS')
               for info in (device_info, valve_info))


def test_find_regex(happi_client, three_valves):
    def find(**kwargs):
        return list(happi_client.backend.find_regex(kwargs))

    valve1 = three_valves['VALVE1']
    valve2 = three_valves['VALVE2']
    valve3 = three_valves['VALVE3']
    assert find(beamline='LCLS') == [valve1, valve2, valve3]
    assert find(beamline='lcls') == [valve1, valve2, valve3]
    assert find(beamline='nomatch') == []
    assert find(_id=r'VALVE\d') == [valve1, valve2, valve3]
    assert find(_id='VALVE[13]') == [valve1, valve3]
    assert find(prefix='BASE:VGC[23]:PV') == [valve2, valve3]


def test_json_delete(mockjson, device_info):
    mockjson.delete(device_info[Client._id_key])
    assert device_info not in mockjson.all_devices


def test_json_save(mockjson, device_info, valve_info):
    # Duplicate device
    with pytest.raises(DuplicateError):
        mockjson.save(device_info[Client._id_key], device_info, insert=True)

    # Device not found
    with pytest.raises(SearchError):
        mockjson.save(valve_info[Client._id_key], valve_info, insert=False)

    # Add to database
    mockjson.save(valve_info[Client._id_key], valve_info, insert=True)
    assert valve_info in mockjson.all_devices


def test_json_locking(mockjson):
    # Place lock on file
    handle = open(mockjson.path, 'w')
    fcntl.flock(handle, fcntl.LOCK_EX | fcntl.LOCK_NB)
    # Attempt to save
    with pytest.raises(IOError):
        mockjson.store({"_ID": "ID"})


def test_json_initialize():
    jb = JSONBackend("testing.json", initialize=True)
    # Check that the file was made
    assert os.path.exists("testing.json")
    # Check it is a valid json file
    assert jb.load() == {}
    # Check that we can not overwrite the database
    with pytest.raises(PermissionError):
        JSONBackend("testing.json", initialize=True)
    # Cleanup
    os.remove("testing.json")


@requires_questionnaire
def test_qs_find(mockqsbackend):
    assert len(list(mockqsbackend.find(dict(beamline='TST')))) == 13
    assert len(list(mockqsbackend.find(dict(name='sam_r')))) == 1


@requires_questionnaire
@requires_pcdsdevices
def test_qsbackend_with_client(mockqsbackend):
    c = Client(database=mockqsbackend)
    assert len(c.all_items) == 13
    assert all(
        d.__class__.__name__ in {'Trigger', 'Motor', 'Acromag'}
        for d in c.all_items
    )
    device_types = [device.__class__.__name__ for device in c.all_items]
    assert device_types.count('Motor') == 6
    assert device_types.count('Trigger') == 2

    # Acromag: six entries, but one erroneously has the same name
    assert device_types.count('Acromag') == 5



class DatabaseError(Exception):
    """Raised when an database intitializes improperly"""
    pass


class EntryError(Exception):
    """Raised when there is an invalid happi entry"""
    pass


class DuplicateError(Exception):
    """Raised when a duplicate device is saved"""
    pass


class ContainerError(Exception):
    """Raised by an improperly setup container"""
    pass


class SearchError(Exception):
    """Raised when no device is found while searching"""
    pass



import collections
import copy
import logging
import re
import sys
from collections import OrderedDict

from prettytable import PrettyTable

from .errors import ContainerError

logger = logging.getLogger(__name__)

# For back-compat to <py3.7
try:
    from re import Pattern
except ImportError:
    from re import _pattern_type as Pattern


class EntryInfo:
    """
    A piece of information related to a specific device

    These are entered as class attributes for a given device container. They
    help control the information entered into a device

    Parameters
    ----------
    doc : str
        A short string to document the device

    optional : bool, optional
        By default all EntryInfo is optional, but in certain cases you may want
        to demand a particular piece of information upon initialization

    enforce : type, list, compiled regex, optional
        Specify that all entered information is entered in a specific format.
        This can either by a Python type i.e. int, float e.t.c., a list of
        acceptable values, or a compiled regex pattern i.e ``re.compile(...)``

    default : optional
        A default value for the trait to have if the user does not specify.
        Keep in mind that this should be the same type as ``enforce`` if you
        are demanding a certain type.

    Raises
    ------
    ContainerError:
        If there is an error with the way the enforced value interacts with its
        default value, or if the piece of information entered is unenforcable
        based on the the settings

    Example
    ------
    .. code::

        class MyDevice(Device):

            my_field = EntryInfo('My generated field')
            number   = EntryInfo('Device number', enforce=int, default=0)
    """
    def __init__(self, doc=None, optional=True, enforce=None, default=None):
        self.key = None  # Set later by parent class
        self.doc = doc
        self.enforce = enforce
        self.optional = optional

        # Explicitly set default to None b/c this is how we ensure mandatory
        # information was set
        if optional:
            self.default = default
        else:
            self.default = None

        # Check that default value is correct type
        try:
            self.enforce_value(default)
        except ValueError:
            raise ContainerError('Default value must match the enforced type')

    def enforce_value(self, value):
        """
        Enforce the rules of the EntryInfo

        Parameters
        ----------
        value

        Returns
        -------
        value :
            Identical to the provided value except it may have been converted
            to the correct type

        Raises
        ------
        ValueError:
            If the value is not the correct type, or does not match the pattern
        """
        if not self.enforce or value is None:
            return value

        elif isinstance(self.enforce, type):
            # Try and convert to type, otherwise raise ValueError
            return self.enforce(value)

        elif isinstance(self.enforce, (list, tuple, set)):
            # Check that value is in list, otherwise raise ValueError
            if value not in self.enforce:
                raise ValueError('{} was not found in the enforce list {}'
                                 ''.format(self.key, self.enforce))
            return value

        elif isinstance(self.enforce, Pattern):
            # Try and match regex patttern, otherwise raise ValueError
            if not self.enforce.match(value):
                raise ValueError(
                    f'{self.key}={value!r} did not match the enforced pattern'
                    f'{self.enforce.pattern}'
                )
            return value

        # Invalid enforcement
        else:
            raise ContainerError('EntryInfo {} has an invalid enforce'
                                 ''.format(self.key))

    def make_docstring(self, parent_class):
        if self.doc is not None:
            return self.doc

        doc = ['{} attribute'.format(self.__class__.__name__),
               '::',
               '',
               ]

        doc.append(repr(self))
        doc.append('')
        return '\n'.join(doc)

    def __get__(self, instance, owner):
        if instance is None:
            return self

        return instance.__dict__[self.key]

    def __set__(self, instance, value):
        instance.__dict__[self.key] = self.enforce_value(value)

    def __repr__(self):
        return 'EntryInfo {} (optional={}, default={})'.format(self.key,
                                                               self.optional,
                                                               self.default)

    def __copy__(self):
        return EntryInfo(doc=self.doc, optional=self.optional,
                         enforce=self.enforce, default=self.default)


class _HappiItemBase:
    def __init_subclass__(cls, **kargs):
        # These attributes are not to be overwritten by subclasses
        RESERVED_ATTRS = [
            '_id', '_info_attrs',
            'creation', 'entry_info', 'info_names', 'last_edit',
            'mandatory_info', 'post', 'save',
            ]

        # Create dict to hold information
        cls._info_attrs = OrderedDict()

        # Handle multiple inheritance
        for base in reversed(cls.mro()[1:]):
            if not hasattr(base, '_info_attrs'):
                continue

            for attr, info in base._info_attrs.items():
                cls._info_attrs[attr] = info

        # Load from highest classEntry
        entries = [(attr, entry)
                   for attr, entry in cls.__dict__.items()
                   if isinstance(entry, EntryInfo)
                   ]

        for attr, entry in entries:
            if attr in RESERVED_ATTRS:
                raise TypeError(
                    f"The attribute name {attr} is used by the HappiItem "
                    f"class and cannot be used as a name for EntryInfo. "
                    f"Choose a different name."
                )

            cls._info_attrs[attr] = entry

        # Notify Info of key names
        for attr, info in cls._info_attrs.items():
            info.key = attr

        # Create docstring information
        for info in cls._info_attrs.values():
            info.__doc__ = info.make_docstring(cls)

        # Store Entry Information
        cls.entry_info = list(cls._info_attrs.values())
        cls.info_names = [info.key for info in cls.entry_info]
        cls.mandatory_info = [info.key for info in cls.entry_info
                              if not info.optional]


class HappiItem(_HappiItemBase, collections.abc.Mapping):
    """
    The smallest description of an object that can be entered in ``happi``

    The class does not need to be intialized with any specific piece of
    information except a name, but all of the attributes listed by
    :attr:`HappiItem.info_names` can be used to assign values to
    :class:`.EntryInfo` upon initialization.  Pieces of information that are
    deemed mandatory by the class must be filled in before the device is loaded
    into the database. See :attr:`.mandatory_info` to see which
    attributes are neccesary.

    Additional metadata can be given to the device in the form of keywords on
    initialization, this information is kept in the :attr:`.extraneous`
    attribute, and will be saved in to the database as long as it does not
    clash with an existing piece of metadata that the client uses to organize
    devices.

    Attributes
    ----------
    entry_info : list
        A list of all the :class:`.EntryInfo` associated with the device

    extraneous : dict
        Storage for information supplied during initialization that does not
        correspond to a specific EntryInfo

    Raises
    ------
    EntryError:
        If a piece of information supplied at startup is of the incorrect type

    ContainerError:
        If one of the pieces of :class:`.EntryInfo` has a default value of the
        incorrect type

    Example
    -------
    .. code ::

        d = HappiItem(name = 'my_device',         # Alias name for device
                      note  = 'Example',          # Piece of arbitrary metadata
                     )
    """
    name = EntryInfo('Shorthand Python-valid name for the Python instance',
                     optional=False,
                     enforce=re.compile(r'[a-z][a-z\_0-9]{2,78}$'))
    device_class = EntryInfo("Python class that represents the instance",
                             enforce=str)
    args = EntryInfo("Arguments to pass to device_class",
                     enforce=list, default=[])
    kwargs = EntryInfo("Keyword arguments to pass to device_class",
                       enforce=dict, default={})
    active = EntryInfo('Whether the instance is actively deployed',
                       enforce=bool, default=True)
    documentation = EntryInfo("Relevant documentation for the instance",
                              enforce=str)

    def __init__(self, **kwargs):
        # Load given information into device class
        for info in self.entry_info:
            if info.key in kwargs.keys():
                setattr(self, info.key, kwargs.pop(info.key))
            else:
                setattr(self, info.key, info.default)
        # Handle additional information
        self.extraneous = kwargs
        if kwargs:
            logger.debug('Additional information for %s was defined %s',
                         self.name, ', '.join(self.extraneous))

    def show_info(self, handle=sys.stdout):
        """
        Show the device instance information in a PrettyTable

        Parameters
        ----------
        handle : file-like, optional
            Option to write to a file-like object
        """
        pt = PrettyTable(['EntryInfo', 'Value'])
        pt.align = 'r'
        pt.align['EntryInfo'] = 'l'
        pt.align['Value'] = 'l'
        pt.float_format = '8.5'

        # Gather all device information, do not show private
        # information that begins with an underscore
        show_info = self.post()
        public_keys = sorted([key for key in show_info.keys()
                              if not key.startswith('_')])
        for key in public_keys:
            pt.add_row([key, show_info[key]])

        print(pt, file=handle)

    def post(self):
        """
        Create a document to be loaded into the happi database

        Returns
        -------
        post : dict
            Dictionary of all contained information
        """
        # Grab all the specified information
        post = dict([(key, getattr(self, key)) for key in self.info_names])

        # Add additional metadata
        if self.extraneous:
            post.update(self.extraneous)

        return post

    def __getitem__(self, item):
        return self.post()[item]

    def __iter__(self):
        yield from self.info_names
        if self.extraneous:
            yield from self.extraneous

    def __len__(self):
        return len(self.post())

    def __copy__(self):
        return self.__deepcopy__({})

    def __deepcopy__(self, memo):
        device_info = copy.deepcopy(dict(self))
        return type(self)(**device_info)

    def save(self):
        """
        Overwritten when the device instance is loaded from the client
        """
        raise NotImplementedError

    def __repr__(self):
        return '{} (name={})'.format(self.__class__.__name__,
                                     self.name)

    def __eq__(self, other):
        return (self.post() == other.post())


class OphydItem(HappiItem):
    prefix = EntryInfo('A base PV for all related records',
                       optional=False, enforce=str)

    args = copy.copy(HappiItem.args)
    args.default = ['{{prefix}}']
    kwargs = copy.copy(HappiItem.kwargs)
    kwargs.default = {'name': '{{name}}'}



import logging
import collections
from qtpy import QtCore, QtGui, QtWidgets

from ..utils import get_happi_entry_value

logger = logging.getLogger(__name__)


class HappiViewMixin(object):
    """
    Base class to be used for View widgets
    """
    def __init__(self, client=None, **kwargs):
        super().__init__(**kwargs)
        self._client = client
        self._entries = []

    @property
    def client(self):
        """
        The client to use for search.

        Returns
        -------
        happi.Client
        """
        return self._client

    @client.setter
    def client(self, client):
        self._client = client

    def entries(self):
        """
        List of search results.

        Returns
        -------
        list
        """
        return self._entries

    def search(self, *args, **kwargs):
        """
        Performs a search into the Happi database and populate the model
        with the new data.

        args and kwargs are sent directly to happi.Client.search method.
        """
        self._entries = self._client.search(*args, **kwargs)

    @staticmethod
    def create_item(entry):
        itm = QtGui.QStandardItem(entry.name)
        itm.setData(entry)
        itm.setFlags(itm.flags() & ~QtCore.Qt.ItemIsEditable)
        return itm


class HappiDeviceListView(QtWidgets.QListView, HappiViewMixin):
    """
    QListView which displays Happi entries.

    Parameters
    ----------
    parent : QWidget
        The parent widget

    client : Client
        A happi.Client instance

    kwargs : dict
        Additional arguments to be passed to the QListView constructor.
    """
    def __init__(self, parent=None, client=None, **kwargs):
        super().__init__(parent=parent, client=client, **kwargs)
        self.model = QtGui.QStandardItemModel()
        self.model.setHorizontalHeaderLabels(["Devices"])

        self.proxy_model = QtCore.QSortFilterProxyModel()
        self.proxy_model.setFilterKeyColumn(-1)
        self.proxy_model.setDynamicSortFilter(True)
        self.proxy_model.setSourceModel(self.model)
        self.setModel(self.proxy_model)

    def search(self, *args, **kwargs):
        """
        Performs a search into the Happi database and populate the model
        with the new data.

        args and kwargs are sent directly to happi.Client.search method.
        """
        super().search(*args, **kwargs)
        self._update_data()

    def _update_data(self):
        """
        Update the model with new data from the search.
        """
        if not self.entries():
            return
        items = [self.create_item(entry.item) for entry in self.entries()]

        self.model.clear()

        for row, itm in enumerate(items):
            self.model.setItem(row, itm)
        self.proxy_model.setSourceModel(self.model)
        self.proxy_model.sort(0, QtCore.Qt.AscendingOrder)


class HappiDeviceTreeView(QtWidgets.QTreeView, HappiViewMixin):
    """
    QListView which displays Happi entries.

    Parameters
    ----------
    parent : QWidget
        The parent widget

    client : Client
        A happi.Client instance

    kwargs : dict
        Additional arguments to be passed to the QListView constructor.
    """
    def __init__(self, parent=None, client=None, **kwargs):
        super().__init__(parent=parent, client=client, **kwargs)
        self.setSortingEnabled(True)
        self._models = dict()
        self._groups = []
        self._active_group = ""

        self.proxy_model = QtCore.QSortFilterProxyModel()
        self.proxy_model.setFilterKeyColumn(-1)
        self.proxy_model.setRecursiveFilteringEnabled(True)
        self.proxy_model.setDynamicSortFilter(True)
        self.setModel(self.proxy_model)

    def search(self, *args, **kwargs):
        """
        Performs a search into the Happi database and populate the model
        with the new data.

        args and kwargs are sent directly to happi.Client.search method.
        """
        super().search(*args, **kwargs)
        self._update_data()

    def group_by(self, field, force=False):
        if field and (self._active_group != field or force):
            self._active_group = field
            model = self._models.get(field, None)
            if not model:
                logger.error('Group model for %s does not exist. Update the '
                             'groups information first.')
            self.proxy_model.setSourceModel(model)

    @property
    def groups(self):
        """
        List of fields to be used when grouping Happi entries.

        Returns
        -------
        list
        """
        return self._groups

    @groups.setter
    def groups(self, groups):
        if self._groups != groups and groups:
            self._groups = groups
            if not self._active_group:
                self._active_group = groups[0]
            self._update_data()

    def _create_group_model(self, field, force=False):
        if field in self._models and not force:
            return

        model = QtGui.QStandardItemModel()
        model.setHorizontalHeaderLabels(["Devices"])

        entry_group = collections.defaultdict(list)

        for entry in self.entries():
            try:
                field_val = get_happi_entry_value(entry.item, field)
                entry_group[field_val].append(entry.item)
            except ValueError:
                logger.exception(
                    'Could not retrieve value for field %s at entry %s',
                    field, entry
                )

        for idx, (key_value, entries) in enumerate(entry_group.items()):
            root = QtGui.QStandardItem(key_value)
            # Disable edit
            root.setFlags(root.flags() & ~QtCore.Qt.ItemIsEditable)

            if len(entries) == 1 and entries[0].name == key_value:
                root.setData(entries[0])
            else:
                # Pack the entries into the root
                root.setData(entries)

                for entry in entries:
                    root.appendRow(self.create_item(entry))
            model.appendRow(root)

        self._models[field] = model
        return model

    def _update_data(self):
        """
        Update the model with new data from the search.
        """
        for field in self._groups:
            if not field:
                return
            self._create_group_model(field, force=True)
        self.group_by(self._active_group, force=True)



import versioneer
from setuptools import (setup, find_packages)

with open('requirements.txt', 'rt') as f:
    requirements = f.read().splitlines()

setup(name='happi',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      author='SLAC National Accelerator Laboratory',
      packages=find_packages(),
      include_package_data=True,
      install_requires=requirements,
      description='Happi Database Access for LCLS Beamline Devices',
      entry_points={'console_scripts': ['happi=happi.cli:main']})



"""
Base backend database options
"""
import logging

logger = logging.getLogger(__name__)


class _Backend:
    """
    Base class for backend database
    """

    @property
    def all_devices(self):
        """
        List of all device sub-dictionaries
        """
        raise NotImplementedError

    def find(self, multiples=False, **kwargs):
        """
        Find an instance or instances that matches the search criteria

        Parameters
        ----------
        multiples : bool
            Find a single result or all results matching the provided
            information

        kwargs :
            Requested information
        """
        raise NotImplementedError

    def save(self, _id, post, insert=True):
        """
        Save information to the database

        Parameters
        ----------
        _id : str
            ID of device

        post : dict
            Information to place in database

        insert : bool, optional
            Whether or not this a new device to the database

        Raises
        ------
        DuplicateError:
            If insert is True, but there is already a device with the provided
            _id

        SearchError:
            If insert is False, but there is no device with the provided _id

        PermissionError:
            If the write operation fails due to issues with permissions
        """
        raise NotImplementedError

    def delete(self, _id):
        """
        Delete a device instance from the database

        Parameters
        ----------
        _id : str
            ID of device

        Raises
        ------
        PermissionError:
            If the write operation fails due to issues with permissions
        """
        raise NotImplementedError



import copy
import logging
import re
import warnings

from .item import EntryInfo, HappiItem

logger = logging.getLogger(__name__)


class Device(HappiItem):
    """
    Deprecated due to confusing naming

    Original docstring:
    A Generic Device container

    Meant for any object will be loaded to represent a physical object in the
    controls system. Contains information on the physical location of the
    device as well as various
    """
    prefix = EntryInfo('A base PV for all related records',
                       optional=False, enforce=str)
    beamline = EntryInfo('Section of beamline the device belongs',
                         optional=False, enforce=str)
    location_group = EntryInfo('Grouping parameter for device location',
                               optional=False, enforce=str)
    functional_group = EntryInfo('Grouping parameter for device function',
                                 optional=False, enforce=str)
    z = EntryInfo('Beamline position of the device',
                  enforce=float, default=-1.0)
    stand = EntryInfo('Acronym for stand, must be three alphanumeric '
                      'characters like an LCLSI stand (e.g. DG3) or follow '
                      'the LCLSII stand naming convention (e.g. L0S04).',
                      enforce=re.compile(r'[A-Z0-9]{3}$|[A-Z][0-9]S[0-9]{2}$'))
    detailed_screen = EntryInfo('The absolute path to the main control screen',
                                enforce=str)
    embedded_screen = EntryInfo('The absolute path to the '
                                'embedded control screen',
                                enforce=str)
    engineering_screen = EntryInfo('The absolute path to '
                                   'the engineering control screen',
                                   enforce=str)
    system = EntryInfo('The system the device is involved with, i.e '
                       'Vacuum, Timing e.t.c',
                       enforce=str)
    macros = EntryInfo("The EDM macro string asscociated with the "
                       "with the device. By using a jinja2 template, "
                       "this can reference other EntryInfo keywords.",
                       enforce=str)
    lightpath = EntryInfo("If the device should be included in the "
                          "LCLS Lightpath", enforce=bool, default=False)
    documentation = EntryInfo("Relevant documentation for the Device",
                              enforce=str)
    parent = EntryInfo('If the device is a component of another, '
                       'enter the name', enforce=str)
    args = copy.copy(HappiItem.args)
    args.default = ['{{prefix}}']
    kwargs = copy.copy(HappiItem.kwargs)
    kwargs.default = {'name': '{{name}}'}

    def __init__(self, *args, **kwargs):
        warnings.warn("happi.device.Device is deprecated. Please use "
                      "OphydItem or LCLSItem.", DeprecationWarning)
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return '{} (name={}, prefix={}, z={})'.format(
                                    self.__class__.__name__,
                                    self.name,
                                    self.prefix,
                                    self.z)

    @property
    def screen(self):
        warnings.warn("The 'screen' keyword is no longer used in Happi as it "
                      "lacks specificity. Use one of detailed_screen, "
                      "embedded_screen, or engineering screen instead",
                      DeprecationWarning)
        return self.detailed_screen



import pytest

from happi import Device, EntryInfo, cache
from happi.loader import fill_template, from_container, load_devices
from happi.utils import create_alias


class TimeDevice(Device):
    days = EntryInfo("Number of days", enforce=int)


def test_fill_template(device):
    # Check that we can properly render a template
    template = "{{name}}"
    assert device.name == fill_template(template, device)
    # Check that we can enforce a type
    template = '{{active}}'
    active = fill_template(template, device, enforce_type=True)
    assert isinstance(active, bool)
    # Check that we will convert a more complex template
    template = '{{name|length()}}'
    text_len = fill_template(template, device, enforce_type=False)
    assert len(device.name) == int(text_len)
    # Check that we can handle non-jinja template
    template = "blah"
    assert template == fill_template(template, device, enforce_type=True)
    # Check that we do not enforce a NoneType
    template = "{{detailed_screen}}"
    assert fill_template(template, device, enforce_type=True) == ""


def test_from_container():
    # Create a datetime device
    d = TimeDevice(name='test', prefix='Tst:This:1', beamline='TST',
                   device_class='datetime.timedelta', args=list(), days=10,
                   kwargs={'days': '{{days}}', 'seconds': 30})
    td = from_container(d)
    # Now import datetime and check that we constructed ours correctly
    import datetime
    assert td == datetime.timedelta(days=10, seconds=30)


def test_caching():
    # Create a datetime device
    d = TimeDevice(name='test', prefix='Tst:This:2', beamline='TST',
                   device_class='types.SimpleNamespace', args=list(), days=10,
                   kwargs={'days': '{{days}}', 'seconds': 30})
    td = from_container(d)
    assert d.name in cache
    assert id(td) == id(from_container(d))
    assert id(td) != id(from_container(d, use_cache=False))
    # Modify md and check we see a reload
    d.days = 12
    assert id(td) != id(from_container(d, use_cache=True))
    # Check with a device where metadata is unavailable
    d = TimeDevice(name='test', prefix='Tst:Delta:3', beamline='TST',
                   device_class='datetime.timedelta', args=list(), days=10,
                   kwargs={'days': '{{days}}', 'seconds': 30})
    td = from_container(d)
    assert id(td) != id(from_container(d))
    assert id(td) != id(from_container(d, use_cache=False))


def test_add_md():
    d = Device(name='test', prefix='Tst:This:3',
               beamline="TST", args=list(),
               device_class="happi.Device")
    obj = from_container(d, attach_md=True)
    assert obj.md.beamline == 'TST'
    assert obj.md.name == 'test'


@pytest.mark.parametrize('threaded', [False, True])
@pytest.mark.parametrize('post_load', [None, lambda x: None])
def test_load_devices(threaded, post_load):
    # Create a bunch of devices to load
    devs = [TimeDevice(name='test_1', prefix='Tst1:This', beamline='TST',
                       device_class='datetime.timedelta', args=list(), days=10,
                       kwargs={'days': '{{days}}', 'seconds': 30}),
            TimeDevice(name='test_2', prefix='Tst2:This', beamline='TST',
                       device_class='datetime.timedelta', args=list(), days=10,
                       kwargs={'days': '{{days}}', 'seconds': 30}),
            TimeDevice(name='test_3', prefix='Tst3:This', beamline='TST',
                       device_class='datetime.timedelta', args=list(), days=10,
                       kwargs={'days': '{{days}}', 'seconds': 30}),
            Device(name='bad', prefix='Not:Here', beamline='BAD',
                   device_class='non.existant')]
    # Load our devices
    space = load_devices(*devs, pprint=True, use_cache=False,
                         threaded=threaded, post_load=post_load)
    # Check all our devices are there
    assert all([create_alias(dev.name) in space.__dict__ for dev in devs])
    # Devices were loading properly or exceptions were stored
    import datetime
    assert space.test_1 == datetime.timedelta(days=10, seconds=30)
    assert isinstance(space.bad, ImportError)



import collections
import configparser
import inspect
import itertools
import logging
import os
import re
import sys
import time as ttime
import warnings

from . import containers
from .backends import BACKENDS, DEFAULT_BACKEND
from .backends.core import _Backend
from .errors import DatabaseError, EntryError, SearchError
from .item import HappiItem
from .loader import from_container

logger = logging.getLogger(__name__)


def _looks_like_database(obj):
    """
    Does the given object look like a backend we can use or does it inherit
    from _Backend
    """
    return (isinstance(obj, _Backend) or
            all(
                hasattr(obj, attr) for attr in (
                    'find', 'all_devices', 'delete', 'save')
                )
            )


class SearchResult(collections.abc.Mapping):
    '''
    A single search result from ``Client.search``

    This result can be keyed for metadata as in::
        result['name']

    The HappiItem can be readily retrieved::
        result.item

    Or the object may be instantiated::
        result.get()

    Attributes
    ----------
    item : happi.HappiItem
        The container
    metadata : dict
        The HappiItem metadata
    '''

    def __init__(self, client, device):
        self._device = device
        self._instantiated = None
        self.client = client
        self.metadata = device.post()

    @property
    def device(self):
        warnings.warn('SearchResult.device deprecated, use SearchResult.item')
        return self.item

    @property
    def item(self):
        if self._device is None:
            self._device = self.client.find_device(**self.metadata)
        return self._device

    def get(self, attach_md=True, use_cache=True, threaded=False):
        '''(get) ''' + from_container.__doc__
        if self._instantiated is None:
            self._instantiated = from_container(
                self.item, attach_md=attach_md, use_cache=use_cache,
                threaded=threaded
            )
        return self._instantiated

    def __getitem__(self, item):
        return self.metadata[item]

    def __iter__(self):
        yield from self.metadata

    def __len__(self):
        return len(self.metadata)

    def __repr__(self):
        return (
            f'{self.__class__.__name__}(client={self.client}, '
            f'metadata={self.metadata})'
        )


class Client(collections.abc.Mapping):
    """
    The client to control the contents of the Happi Database

    Parameters
    ----------
    database : happi.backends.Backend
        A already instantiated backend

    kwargs:
        Passed to the `db_type` backend

    Raises
    -----
    DatabaseError:
        Raised if the Client fails to instantiate the Database
    """
    # HappiItem information
    _client_attrs = ['_id', 'type', 'creation', 'last_edit']
    _id_key = 'name'
    _results_wrap_class = SearchResult

    def __init__(self, database=None, **kwargs):
        # Use supplied backend
        if database:
            self.backend = database
            if not _looks_like_database(database):
                raise ValueError(f'{database!r} does not look like a database;'
                                 f' expecting an instantiated happi backend')
        # Load database
        else:
            logger.debug("No database given, using '%s'", DEFAULT_BACKEND)
            try:
                self.backend = DEFAULT_BACKEND(**kwargs)
            except Exception as exc:
                raise DatabaseError(
                    f"Failed to instantiate a {DEFAULT_BACKEND} backend"
                ) from exc

    def find_document(self, **kwargs):
        """
        Load a device document from the database

        If multiple matches are found, a single document will be returned to
        the user. How the database will choose to select this device is based
        on each individual implementation

        Parameters
        ----------
        kwargs :
            Add information to locate the device in keyword pairs

        Returns
        -------
        document : dict
            A dict that matches the specified information.

        Raises
        ------
        SearchError:
            If no document with the given information is found

        See Also
        --------
        :meth:`.find_device`, :meth:`.search`
        """
        if len(kwargs) == 0:
            raise SearchError('No information pertinent to device given')

        matches = list(itertools.islice(self.backend.find(kwargs), 1))
        if not matches:
            raise SearchError(
                'No device information found that matches the search criteria')
        return matches[0]

    def create_device(self, device_cls, **kwargs):
        """
        Create a new device

        Parameters
        ----------
        device_cls : :class:`.HappiItem` or name of class
            The Device HappiItem to instantiate

        kwargs :
            Information to pass through to the device, upon initialization

        Returns
        -------
        device :
            An instantiated version of the device

        Raises
        ------
        TypeError:
            If the provided class is not a subclass of :class:`.HappiItem`


        Example
        -------
        .. code::

            device = client.create_device(Device,   name='my_device' ...)
            device = client.create_device('Device', name='my_device',...)

        """
        # If specified by a string
        if device_cls in containers.registry:
            device_cls = containers.registry[device_cls]

        # Check that this is a valid HappiItem
        if isinstance(device_cls, str):
            raise TypeError(
                f'The device class {device_cls!r} is not in the registry'
            )

        if (not inspect.isclass(device_cls) or
                not issubclass(device_cls, HappiItem)):
            raise TypeError(f'{device_cls!r} is not a subclass of HappiItem')

        device = device_cls(**kwargs)
        # Add the method to the device

        def save_device():
            self.add_device(device)

        device.save = save_device
        return device

    def add_device(self, device):
        """
        Add a new device into the database

        Parameters
        ----------
        device : :class:`.HappiItem`
            The device to store in the database

        Raises
        ------
        EntryError:
            If all of the mandatory information for the device has not been
            specified or there is already a device with that ``id`` in the
            database
        """
        logger.info("Storing device %r ...", device)
        _id = self._store(device, insert=True)
        logger.info('HappiItem %r has been succesfully added to the '
                    'database', device)

        def save_device():
            self._store(device, insert=False)

        device.save = save_device
        return _id

    def find_device(self, **post):
        """
        Used to query the database for an individual HappiItem

        If multiple devices are found, only the first is returned

        Parameters
        ----------
        post :
            Information to pertinent to the device

        Raises
        ------
        SearchError
            If no match for the given information is found

        Returns
        -------
        device : :class:`.HappiItem`
            A device that matches the characteristics given
        """
        logger.debug("Gathering information about the device ...")
        doc = self.find_document(**post)
        # Instantiate HappiItem
        logger.debug("Instantiating device based on found information ...")
        try:
            device = self.create_device(doc['type'], **doc)
        except (KeyError, TypeError) as exc:
            raise EntryError('The information relating to the device class '
                             'has been modified to the point where the object '
                             'can not be initialized, please load the '
                             'corresponding document') from exc

        # Add the save method to the device
        device.save = lambda: self._store(device, insert=False)
        return device

    def load_device(self, use_cache=True, **post):
        """
        Find a device in the database and instantiate it

        Essentially a shortcut for:

        .. code:: python

            container = client.find_device(name='...')
            device = from_container(container)

        Parameters
        ----------
        post:
            Passed to :meth:`.Client.find_device`

        use_cache: bool, optional
            Choice to use a cached device. See :meth:`.from_container` for more
            information

        Returns
        -------
        obj:
            Instantiated object
        """
        cntr = self.find_device(**post)
        return from_container(cntr, use_cache=use_cache)

    def validate(self):
        """
        Validate all of the devices in the database by attempting to initialize
        them and asserting their mandatory information is present. Information
        is written to the logger

        Returns
        -------
        ids : list
            List of device ids that have failed verification
        """
        bad = list()
        logger.debug('Loading database to validate contained devices ...')
        for post in self.backend.all_devices:
            # Try and load device based on database info
            try:
                # HappiItem identification
                _id = post[self._id_key]
                logger.debug('Attempting to initialize %s...', _id)
                # Load HappiItem
                device = self.find_device(**post)
                logger.debug('Attempting to validate ...')
                self._validate_device(device)
            except KeyError:
                logger.error("Post has no id  %s", post)
            # Log all generated exceptions
            except Exception as e:
                logger.warning("Failed to validate %s because %s", _id, e)
                bad.append(_id)
            # Report successes
            else:
                logger.debug('Successfully validated %s', _id)
        return bad

    @property
    def all_devices(self):
        """
        A list of all contained devices
        """
        warnings.warn('Client.all_devices deprecated, use all_items')
        return self.all_items

    @property
    def all_items(self):
        """
        A list of all contained devices
        """
        return [res.item for res in self.search()]

    def __getitem__(self, key):
        'Get a device ID'
        try:
            device = self.find_device(**self.backend.get_by_id(key))
        except Exception as ex:
            raise KeyError(key) from ex

        return SearchResult(client=self, device=device)

    def __iter__(self):
        for info in self.backend.find({}):
            yield info['_id']

    def __len__(self):
        return len(self.all_items)

    def _get_search_results(self, items, *, wrap_cls=None):
        '''
        Return search results to the user, optionally wrapping with a class
        '''
        wrap_cls = wrap_cls or self._results_wrap_class
        results = []
        for info in items:
            try:
                result = wrap_cls(client=self, device=self.find_device(**info))
                results.append(result)
            except Exception as exc:
                logger.warning('Entry for %s is malformed (%s). Skipping.',
                               info['name'], exc)
                raise
        return results

    def search_range(self, key, start, end=None, **kwargs):
        """
        Search the database for a device or devices

        Parameters
        -----------
        key : str
            Database key to search

        start : float, optional
            Minimum beamline position to include devices

        end : float, optional
            Maximum beamline position to include devices

        **kwargs
            Information to filter through the database structured as key, value
            pairs for the desired pieces of EntryInfo

        Returns
        -------
        Either a list of devices or dictionaries

        Example
        -------
        .. code::

            gate_valves = client.search_range('z', 0, 100, type='Valve')
            hxr_valves  = client.search_range('z', 0, 100, type='Valve',
                                              beamline='HXR')
        """
        items = self.backend.find_range(key, start=start, stop=end,
                                        to_match=kwargs)
        return self._get_search_results(items)

    def search(self, **kwargs):
        """
        Search the database for a device or devices

        Parameters
        -----------
        **kwargs
            Information to filter through the database structured as key, value
            pairs for the desired pieces of EntryInfo

        Returns
        -------
        res : list of SearchResult
            The search results

        Example
        -------
        .. code::

            gate_valves = client.search(type='Valve')
            hxr_valves  = client.search(type='Valve', beamline='HXR')
        """
        items = self.backend.find(kwargs)
        return self._get_search_results(items)

    def search_regex(self, flags=re.IGNORECASE, **kwargs):
        """
        Search the database for a device or devices

        Parameters
        -----------
        flags : int, optional
            Defaulting to ``re.IGNORECASE``, these flags are used for the
            regular expressions passed in.

        **kwargs
            Information to filter through the database structured as key, value
            pairs for the desired pieces of EntryInfo.  Every value is allowed
            to contain a Python regular expression.

        Returns
        -------
        Either a list of devices or dictionaries

        Example
        -------
        .. code::

            gate_valves = client.search_regex(beamline='Valve.*')
            three_valves = client.search_regex(_id='VALVE[123]')
        """
        items = self.backend.find_regex(kwargs, flags=flags)
        return self._get_search_results(items)

    def export(self, path=sys.stdout, sep='\t', attrs=None):
        """
        Export the contents of the database into a text file

        Parameters
        ----------
        path : File Handle
            File-like object to save text file

        sep : str
            Separator to place inbetween columns of information

        attrs : iterable
            Attributes to include, these will be a list of values
        """
        # Load documents
        devs = self.all_items
        logger.info('Creating file at %s ...', path)
        # Load device information
        with path as f:
            for dev in devs:
                try:
                    f.write(sep.join([getattr(dev, attr)
                                      for attr in attrs])
                            + '\n')
                except KeyError as e:
                    logger.error("HappiItem %s was missing attribute %s",
                                 dev.name, e)

    def remove_device(self, device):
        """
        Remove a device from the database

        Parameters
        ----------
        device : :class:`.HappiItem`
            HappiItem to be removed from the database
        """
        # HappiItem Check
        if not isinstance(device, HappiItem):
            raise ValueError("Must supply an object of type `HappiItem`")
        logger.info("Attempting to remove %r from the "
                    "collection ...", device)
        # Check that device is in the database
        _id = getattr(device, self._id_key)
        self.backend.delete(_id)

    def _validate_device(self, device):
        """
        Validate that a device has all of the mandatory information
        """
        logger.debug('Validating device %r ...', device)

        # Check type
        if not isinstance(device, HappiItem):
            raise ValueError('{!r} is not a subclass of '
                             'HappiItem'.format(device))
        logger.debug('Checking mandatory information has been entered ...')
        # Check that all mandatory info has been entered
        missing = [info.key for info in device.entry_info
                   if not info.optional and
                   info.default == getattr(device, info.key)]
        # Abort initialization if missing mandatory info
        if missing:
            raise EntryError('Missing mandatory information ({}) for {}'
                             ''.format(', '.join(missing),
                                       device.__class__.__name__))
        logger.debug('HappiItem %r has been validated.', device)

    def _store(self, device, insert=False):
        """
        Store a document in the database

        Parameters
        ----------
        post : :class:`.HappiItem`
            HappiItem to save

        insert : bool, optional
            Set to True if this is a new entry

        Raises
        ------
        DuplicateError:
            When _id already exists

        EntryError:
            If the device doesn't the correct information

        Todo
        ----
        Enforce parent is an already entered name
        """
        logger.debug('Loading a device into the collection ...')

        # Validate device is ready for storage
        self._validate_device(device)
        # Grab information from device
        post = device.post()
        # save the old name in case the user is trying to
        # change the 'name' of an entry
        the_old_name = post.get('_id', None)
        # Store creation time
        creation = post.get('creation', ttime.ctime())
        # Clean supplied information
        [post.pop(key) for key in self._client_attrs if key in post]
        # Note that device has some unrecognized metadata
        for key in post.keys():
            if key not in device.info_names:
                logger.debug("HappiItem %r defines an extra piece of "
                             "information under the keyword %s",
                             device, key)
        # Add metadata from the Client Side

        tpe = containers.registry.entry_for_class(device.__class__)
        post.update({'type': tpe,
                     'creation': creation,
                     'last_edit': ttime.ctime()})

        # Find id
        try:
            _id = post[self._id_key]
        except KeyError:
            raise EntryError('HappiItem did not supply the proper information '
                             'to interface with the database, missing {}'
                             ''.format(self._id_key))
        # In case we want to update the name of an entry
        # We want to add a new entry, and delete the old one
        if the_old_name and the_old_name != post[self._id_key]:
            # Store information for the new entry
            logger.info('Saving new entry %s ...', _id)
            self.backend.save(_id, post, insert=True)
            # Remove the information for the old entry
            logger.info('Removing old entry %s ...', the_old_name)
            self.backend.delete(the_old_name)
        else:
            # Store information
            logger.info('Adding / Modifying information for %s ...', _id)
            self.backend.save(_id, post, insert=insert)
        return _id

    @classmethod
    def from_config(cls, cfg=None):
        """
        Create a client from a configuration file specification

        Configuration files looking something along the lines of:

        .. code::

            [DEFAULT]
            path=path/to/my/db.json

        All key value pairs will be passed directly into backend construction
        with the exception of the key ``backend`` which can be used to specify
        a specific type of backend if this differs from the configured default.

        Parameters
        ----------
        cfg: str, optional
            Path to a configuration file. If not entered, :meth:`.find_config`
            will be use.
        """
        # Find a configuration file
        if not cfg:
            cfg = cls.find_config()

        if not os.path.exists(cfg):
            raise RuntimeError(f'happi configuration file not found: {cfg!r}')

        # Parse configuration file
        cfg_parser = configparser.ConfigParser()
        cfg_file = cfg_parser.read(cfg)
        logger.debug("Loading configuration file at %r", cfg_file)
        db_kwargs = cfg_parser['DEFAULT']
        # If a backend is specified use it, otherwise default
        if 'backend' in db_kwargs:
            db_str = db_kwargs.pop('backend')
            try:
                backend = BACKENDS[db_str]
            except KeyError:
                raise RuntimeError(
                    f'Happi backend {db_str!r} unavailable'
                ) from None
        else:
            backend = DEFAULT_BACKEND

        logger.debug("Using Happi backend %r with kwargs", backend, db_kwargs)
        # Create our database with provided kwargs
        try:
            database = backend(**db_kwargs)
            return cls(database=database)
        except Exception as ex:
            raise RuntimeError(
                f'Unable to instantiate the client. Please verify that '
                f'your HAPPI_CFG points to the correct file and has '
                f'the required configuration settings. In {cfg!r}, found '
                f'settings: {dict(db_kwargs)}.'
            ) from ex

    @staticmethod
    def find_config():
        """
        Search for a ``happi`` configuration file

        We first query the environment variable ``$HAPPI_CFG`` to see if this
        points to a specific configuration file. If this is not present, the
        variable set by ``$XDG_CONFIG_HOME`` or if  that is not set
        ``~/.config``

        Returns
        -------
        path: str
            Absolute path to configuration file

        Raises
        ------
        EnvironmentError:
            If no configuration file can be found by the methodology detailed
            above
        """
        # Point to with an environment variable
        if os.environ.get('HAPPI_CFG', False):
            happi_cfg = os.environ.get('HAPPI_CFG')
            logger.debug("Found $HAPPI_CFG specification for Client "
                         "configuration at %s", happi_cfg)
            return happi_cfg
        # Search in the current directory and home directory
        else:
            config_dir = (os.environ.get('XDG_CONFIG_HOME')
                          or os.path.expanduser('~/.config'))
            logger.debug('Searching for Happi config in %s', config_dir)
            for path in ('.happi.cfg', 'happi.cfg'):
                full_path = os.path.join(config_dir, path)
                if os.path.exists(full_path):
                    logger.debug("Found configuration file at %r", full_path)
                    return full_path
        # If found nothing
        raise EnvironmentError("No happi configuration file found. "
                               "Check HAPPI_CFG.")

    def choices_for_field(self, field):
        """
        List all choices for a given field

        Parameters
        ----------
        field : string
            search field to list all possible choices for
            i.e 'beamline', 'name', 'z', 'prefix', etc.

        Raises
        ------
        SearchError
            If no devices in the database have an entry for the given field

        Returns
        -------
        field_choices : list
            list of choices for a given field that are in the database
        """
        field_choices = set()
        for dev in self.all_items:
            try:  # Want to ignore error if 'dev' doesn't have 'field'
                choice = getattr(dev, field)
                field_choices.add(choice)
            except AttributeError:
                pass
        if len(field_choices) == 0:
            raise SearchError('No entries found with given field')
        return field_choices



__all__ = ['Device', 'EntryInfo', 'Client', 'from_container', 'load_devices',
           'cache', 'HappiItem', 'OphydItem', 'SearchResult']
from ._version import get_versions
from .client import Client, SearchResult
from .device import Device
from .item import EntryInfo, HappiItem, OphydItem
from .loader import cache, from_container, load_devices

__version__ = get_versions()['version']
del get_versions
